from biosiglive import StreamData, TcpClient, DeviceType, RealTimeProcessingMethod, InverseKinematicsMethods
from examples.custom_interface import MyInterface
from threading import Thread
import os
import time


# TODO fix this test so the program end when it is done
def test_tcp_client():
    parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    server_ip = "127.0.0.1"
    server_port = 50000
    model_path = parent_dir + "/examples/model/Wu_Shoulder_Model_mod_wt_wrapp.bioMod"
    nb_electrode = 5
    interface = MyInterface(system_rate=100, data_path=parent_dir + "/examples/abd.bio")
    interface.add_device(
        name="EMG",
        device_type=DeviceType.Emg,
        rate=2000,
        nb_channels=nb_electrode,
        device_data_file_key="emg",
        data_buffer_size=2000,
        processing_method=RealTimeProcessingMethod.ProcessEmg,
        processing_window=1,
        moving_average_window=1,
        low_pass_filter=False,
        band_pass_filter=True,
        normalization=False,
    )

    interface.add_marker_set(
        name="markers",
        rate=100,
        data_buffer_size=100,
        kinematics_method=InverseKinematicsMethods.BiorbdKalman,
        model_path=model_path,
        marker_data_file_key="markers",
        nb_markers=16,
        unit="mm",
    )

    def stream_data(interface):
        stream_data = StreamData(stream_rate=100)
        stream_data.add_interface(interface)
        stream_data.add_server(server_ip, server_port)
        stream_data.start()

    server_thread = Thread(target=stream_data, args=(interface,))
    server_thread.start()
    time.sleep(1)
    # create a client
    tcp_client = TcpClient(server_ip, server_port, read_frequency=100)
    i = 0
    command = ["proc_device_data", "marker_set_data", "raw_device_data", "kinematics_data"]
    data = {}
    while i != 5:
        data = tcp_client.get_data_from_server(command=command, nb_frame_to_get=2, down_sampling={"emg": 20})
        i += 1

    shapes_0 = [5, 3, 5, 15]
    shapes_1 = [2, 16, 2, 2]
    for c, com in enumerate(command):
        assert com == list(data.keys())[c]
        assert data[com].shape[0] == shapes_0[c]
        assert data[com].shape[1] == shapes_1[c]
        if len(data[com].shape) == 3:
            assert data[com].shape[2] == 1



import pickle
import numpy as np
from pathlib import Path


def save(data_dict, data_path):
    """This function adds data to a pickle file. It not open the file, but appends the data to the end, so it's fast.

    Parameters
    ----------
    data_dict : dict
        The data to be added to the file.
    data_path : str
        The path to the file. The file must exist.
    """
    if Path(data_path).suffix != ".bio":
        if Path(data_path).suffix == "":
            data_path += ".bio"
        else:
            raise ValueError("The file must be a .bio file.")
    with open(data_path, "ab") as outf:
        pickle.dump(data_dict, outf, pickle.HIGHEST_PROTOCOL)


# TODO add dict merger
def load(filename, number_of_line=None):
    """This function reads data from a pickle file to concatenate them into one dictionary.

    Parameters
    ----------
    filename : str
        The path to the file.
    number_of_line : int
        The number of lines to read. If None, all lines are read.

    Returns
    -------
    data : dict
        The data read from the file.

    """
    if Path(filename).suffix != ".bio":
        raise ValueError("The file must be a .bio file.")
    data = {}
    limit = 2 if not number_of_line else number_of_line
    with open(filename, "rb") as file:
        count = 0
        while count < limit:
            try:
                data_tmp = pickle.load(file)
                for key in data_tmp.keys():
                    if key in data.keys():
                        if isinstance(data[key], list) is True:
                            data[key].append(data_tmp[key])
                        else:
                            data[key] = np.append(data[key], data_tmp[key], axis=len(data[key].shape) - 1)
                    else:
                        if isinstance(data_tmp[key], (int, float, str, dict)) is True:
                            data[key] = [data_tmp[key]]
                        elif isinstance(data_tmp[key], list) is True:
                            data[key] = [data_tmp[key]]
                        else:
                            data[key] = data_tmp[key]
                if number_of_line:
                    count += 1
                else:
                    count = 1
            except EOFError:
                break
    return data



"""
This example shows how to start a server to send data over a TCP/IP connection.
The server is first created with the IP address and port as input, and then it is started. In the loop, the server waits for a connection from a client. If a client sends a message to the server, the server reads the message and sends data based on the connection that was just established.
To try this example, please run this code, then the get_from_server.py example.
"""

from biosiglive import Server
import numpy as np

if __name__ == "__main__":
    server_ip = "127.0.0.1"
    port = 50000
    server = Server(server_ip, port)
    server.start()
    while True:
        print("Server is listening")
        connection, message = server.client_listening()
        data = {"proc_device_data": np.random.rand(5, 100), "marker_set_data": np.random.rand(15, 100)}
        server.send_data(data, connection, message)
        print("Data sent to the client.")



from biosiglive import OfflinePlot
import numpy as np
import matplotlib

matplotlib.use("Agg")


def test_offline_plot():
    np.random.seed(50)
    data = np.random.random((5, 20))
    OfflinePlot.multi_plot(data)


# TODO add test live plot



import pytest
from biosiglive import ViconClient, DeviceType
import numpy as np


@pytest.mark.parametrize("init", [True, False])
@pytest.mark.parametrize("get_frame", [True, False])
def test_interface(init, get_frame):
    # Create interface
    interface = ViconClient(system_rate=100, init_now=init)

    # Add device
    interface.add_device(
        name="EMG",
        device_type=DeviceType.Emg,
        nb_channels=5,
    )
    interface.add_marker_set(nb_markers=5, name="markers", rate=100)

    if not init:
        interface.init_client()
    i = 0
    while i != 50:
        # Get data
        if not get_frame:
            interface.get_frame()
        device_data = interface.get_device_data(get_frame=get_frame)
        markers_data = interface.get_marker_set_data(get_frame=get_frame)
        latency = [interface.get_latency()]
        frame_number = [interface.get_frame_number()]

        i += 1

    np.testing.assert_almost_equal(device_data.shape[0], 5)
    np.testing.assert_almost_equal(markers_data.shape[1], 10)
    np.testing.assert_almost_equal(len(latency), 1)
    np.testing.assert_almost_equal(len(frame_number), 1)



"""
This file define a wrapper for the socket client to receive data from the server.
"""

import socket
import struct
from typing import Union
import pickle

Buff_size = 32767


class Message:
    def __init__(
        self,
        command: Union[list, str] = "all",
        nb_frame_to_get: int = 1,
        down_sampling: dict = None,
        custom_cmd: Union[str, list] = None,
    ):
        """
        Message class

        Parameters
        ----------
        command: Union[list, str]
            List of commands to send to the server.
        nb_frame_to_get: int
            Number of frames to get from the server.
        down_sampling: dict
            Dictionary containing the down sampling number for data in command.
        """
        if isinstance(command, str):
            command = [command]
        self.command = command
        self.nb_frames_to_get = nb_frame_to_get
        self.down_sampling = down_sampling
        self.custom_cmd = custom_cmd


class Client:
    def __init__(self, server_ip: str, port: int, client_type: str = "TCP", name: str = None):
        """
        Create a client main.

        Parameters
        ----------
        server_ip: str
            Server address.
        port: int
            Server port.
        client_type: str
            Type of the main.
        name: str
            Name of the client.
        """

        self.name = name if name is not None else "Client"
        self.client_type = client_type
        self.address = f"{server_ip}:{port}"
        self.server_address = server_ip
        self.port = port
        self.client = self.client_sock(self.client_type)

    def _connect(self):
        self.client = self.client_sock(self.client_type)
        self.client.connect((self.server_address, self.port))

    @staticmethod
    def client_sock(
        tcp_type: str,
    ):
        """
        Create a client main.

        Parameters
        ----------
        tcp_type: str
            Type of the main.

        Returns
        -------
        client: socket.socket
            Client main.
        """
        if tcp_type == "TCP" or type is None:
            return socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        elif tcp_type == "UDP":
            return socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        else:
            raise RuntimeError(f"Invalid type of connexion ({tcp_type}). Type must be 'TCP' or 'UDP'.")

    def _recv_all(self, buff_size: int = Buff_size):
        """
        Receive all data from the server.

        Parameters
        ----------
        buff_size: int
            Size of the buffer.

        Returns
        -------
        data: list
            List of data received.
        """

        msg_len = self.client.recv(4)
        msg_len = struct.unpack(">I", msg_len)[0]
        data = []
        l = 0
        while l < msg_len:
            chunk = self.client.recv(buff_size)
            l += len(chunk)
            data.append(chunk)
        data = b"".join(data)
        data = pickle.loads(data)
        return data

    def get_data(self, message: Message = Message(), buff: int = Buff_size):
        """
        Get the data from server using the command.

        Parameters
        ----------
        message: Message
            Message to send to the server.
        buff: int
            Size of the buffer.

        Returns
        -------
        data: dict
            Data from server.
        """
        message = message.__dict__
        self._connect()
        self.client.sendall(pickle.dumps(message))
        return self._recv_all(buff)



import pytest
from biosiglive import (
    RealTimeProcessing,
    OfflineProcessing,
    RealTimeProcessingMethod,
    OfflineProcessingMethod,
    MskFunctions,
    InverseKinematicsMethods,
)
import numpy as np
import os


def custom_function(new_sample):
    return new_sample


@pytest.mark.parametrize(
    "method",
    [
        RealTimeProcessingMethod.CalibrationMatrix,
        RealTimeProcessingMethod.ProcessEmg,
        RealTimeProcessingMethod.GetPeaks,
        RealTimeProcessingMethod.ProcessGenericSignal,
        RealTimeProcessingMethod.ProcessImu,
        RealTimeProcessingMethod.Custom,
    ],
)
def test_real_time_processing(method):
    np.random.seed(50)
    data = np.random.rand(2, 20)
    data_3d = np.random.rand(4, 3, 20)
    processing = RealTimeProcessing(data_rate=2000, processing_window=1000)
    processed_data = None
    nb_peaks = None
    i = 0
    while i != 150:
        if method == RealTimeProcessingMethod.CalibrationMatrix:
            processed_data = processing.calibration_matrix(data, np.random.rand(2, 2))
        elif method == RealTimeProcessingMethod.ProcessEmg:
            processed_data = processing.process_emg(data)
        elif method == RealTimeProcessingMethod.GetPeaks:
            data_tmp = processing.get_peaks(data, threshold=0.1)
            processed_data = data_tmp[1]
            nb_peaks = data_tmp[0]
        elif method == RealTimeProcessingMethod.ProcessGenericSignal:
            processed_data = processing.process_generic_signal(data)
        elif method == RealTimeProcessingMethod.ProcessImu:
            processed_data = processing.process_imu(data_3d)
        elif method == RealTimeProcessingMethod.Custom:
            processed_data = processing.custom_processing(custom_function, data)

        if method == RealTimeProcessingMethod.ProcessImu:
            np.testing.assert_almost_equal(len(processed_data.shape), 3)
        else:
            np.testing.assert_almost_equal(len(processed_data.shape), 2)
        i += 1

    if method == RealTimeProcessingMethod.CalibrationMatrix:
        np.testing.assert_almost_equal(processed_data[:, -1], [0.4959683, 0.3568878])
    elif method == RealTimeProcessingMethod.ProcessEmg:
        np.testing.assert_almost_equal(processed_data[:, -1], [0.156469, 0.1365517])
    elif method == RealTimeProcessingMethod.GetPeaks:
        np.testing.assert_almost_equal(processed_data[:, -1], [0.0, 0.0])
        np.testing.assert_almost_equal(nb_peaks, 100)
    elif method == RealTimeProcessingMethod.ProcessGenericSignal:
        np.testing.assert_almost_equal(processed_data[:, -1], [0.0164153, 0.0218168])
    elif method == RealTimeProcessingMethod.Custom:
        np.testing.assert_almost_equal(processed_data[:, -1], [0.3910874, 0.2220394])
    elif method == RealTimeProcessingMethod.ProcessImu:
        np.testing.assert_almost_equal(processed_data[:, 0, -1], [0.5367742, 0.477345, 0.3982464, 0.464923])


@pytest.mark.parametrize("method", [OfflineProcessingMethod.ProcessEmg, OfflineProcessingMethod.ComputeMvc])
def test_offline_processing(method):
    np.random.seed(50)
    data = np.random.rand(2, 4000)
    processing = OfflineProcessing(data_rate=2000, processing_window=1000)

    if method == OfflineProcessingMethod.ProcessEmg:
        processed_data = processing.process_emg(data)
        np.testing.assert_almost_equal(processed_data[:, 0], [0.1021941, 0.0959892])
    elif method == OfflineProcessingMethod.ComputeMvc:
        processed_data = processing.compute_mvc(2, data, 2000)
        np.testing.assert_almost_equal(processed_data[:], [0.8815387, 0.873678])


@pytest.mark.parametrize(
    "methods",
    [
        InverseKinematicsMethods.BiorbdLeastSquare,
        InverseKinematicsMethods.BiorbdKalman,
    ],
)
def test_inverse_kinematics_methods(methods):
    parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    markers_data = np.ones((3, 16, 1)) * 0.1
    model_path = parent_dir + "/examples/model/Wu_Shoulder_Model_mod_wt_wrapp.bioMod"

    msk_function = MskFunctions(model=model_path)
    q, q_dot = None, None
    i = 0
    while i != 15:
        q, q_dot = msk_function.compute_inverse_kinematics(markers_data, method=methods)
        i += 1

    if methods == InverseKinematicsMethods.BiorbdLeastSquare:
        np.testing.assert_almost_equal(
            q[:, 0],
            [
                0.0859061,
                0.1650248,
                0.1019853,
                -2.3876092,
                1.3106894,
                2.7743532,
                -3.8583198,
                -0.2704209,
                2.0373899,
                0.1503456,
                0.9030997,
                0.8522711,
                -1.6218697,
                0.4328726,
                -3.10159,
            ],
            decimal=4,
        )
        np.testing.assert_almost_equal(q_dot[:, 0], np.zeros((15,)))

    if methods == InverseKinematicsMethods.BiorbdKalman:
        np.testing.assert_almost_equal(
            q[:, 0],
            [
                0.1641379,
                0.1138498,
                0.1144206,
                1.2445316,
                0.0922355,
                -0.9251613,
                -3.9271956,
                -2.9052048,
                -12.9825494,
                -2.8963511,
                -14.2619434,
                4.0001431,
                -4.6817521,
                4.2250031,
                -3.1007907,
            ],
            decimal=4,
        )
        np.testing.assert_almost_equal(
            q_dot[:, 0],
            [
                -0.013688,
                -0.0174,
                0.054321,
                3.4537,
                0.611992,
                0.115344,
                -5.570652,
                -3.603674,
                0.370434,
                -1.234856,
                -0.491976,
                -0.033802,
                -0.031005,
                0.610965,
                0.03146,
            ],
            decimal=4,
        )


def test_forward_kinematics():
    parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    np.random.seed(50)
    q_data = np.random.rand(15, 1)
    model_path = parent_dir + "/examples/model/Wu_Shoulder_Model_mod_wt_wrapp.bioMod"

    msk_function = MskFunctions(model=model_path)
    markers = None
    i = 0
    while i != 15:
        markers = msk_function.compute_direct_kinematics(q_data)
        i += 1

    np.testing.assert_almost_equal(
        markers[0, :, 0],
        [
            0.5047011,
            0.6690358,
            0.4138976,
            0.58448,
            0.5091725,
            0.4676433,
            0.6165203,
            0.4549606,
            0.5009109,
            0.7444251,
            0.787134,
            0.5910698,
            0.6242104,
            0.9334877,
            0.7964393,
            0.8911985,
        ],
    )



import pytest
from biosiglive import (
    InterfaceType,
    ViconClient,
    PytrignoClient,
    RealTimeProcessingMethod,
    TcpClient,
    Device,
    DeviceType,
    MarkerSet,
    InverseKinematicsMethods,
)
from examples.custom_interface import MyInterface
import numpy as np
import os


@pytest.mark.parametrize(
    "interface_type",
    [InterfaceType.Custom, InterfaceType.ViconClient, InterfaceType.PytrignoClient, InterfaceType.TcpClient],
)
def test_interface(interface_type):
    # Create interface
    interface = None
    if interface_type == InterfaceType.Custom:
        interface = MyInterface(system_rate=100)
    elif interface_type == InterfaceType.ViconClient:
        interface = ViconClient(system_rate=100, init_now=False)
    elif interface_type == InterfaceType.PytrignoClient:
        interface = PytrignoClient(system_rate=100, ip="127.0.0.1", init_now=False)
    elif interface_type == InterfaceType.TcpClient:
        interface = TcpClient(read_frequency=100, ip="127.0.0.1")

    # Add device
    interface.add_device(
        name="EMG",
        device_type=DeviceType.Emg,
        rate=2000,
        nb_channels=5,
        device_data_file_key="emg",
        data_buffer_size=2000,
        processing_method=RealTimeProcessingMethod.ProcessEmg,
        processing_window=2000,
        low_pass_filter=True,
        band_pass_filter=True,
    )
    np.testing.assert_almost_equal(len(interface.devices), 1)
    assert interface.get_device("EMG") == interface.devices[0]
    assert interface.get_device(idx=0) == interface.devices[0]

    try:
        interface.add_marker_set(name="markers", rate=100, data_buffer_size=100)
    except Exception as e:
        if interface_type == InterfaceType.PytrignoClient:
            assert e
    else:
        np.testing.assert_almost_equal(len(interface.marker_sets), 1)
        assert interface.get_marker_set("markers") == interface.marker_sets[0]
        assert interface.get_marker_set(idx=0) == interface.marker_sets[0]


@pytest.mark.parametrize("device_type", [DeviceType.Emg, DeviceType.Imu, DeviceType.Generic])
def test_devices(device_type):
    device = Device(
        device_type=device_type,
        nb_channels=3,
        name="my_device",
        rate=2000,
        system_rate=100,
        channel_names=["1", "2", "3"],
    )
    device.data_window = 100
    assert device.device_type == device_type
    assert device.sample == 2000 / 100
    assert device.name == "my_device"
    assert device.nb_channels == 3

    np.random.seed(50)
    i = 0
    while i != 100:
        if device_type != DeviceType.Imu:
            device.new_data = np.random.random((3, 20))
        else:
            device.new_data = np.random.random((3, 9, 20))
        device.append_data(device.new_data)
        i += 1

    raw_data = device.raw_data
    if device_type != DeviceType.Imu:
        processed_data = device.process(RealTimeProcessingMethod.ProcessGenericSignal)
        assert raw_data.shape == (3, 100)
        assert processed_data.shape == (3, 20)
    else:
        processed_data = device.process(RealTimeProcessingMethod.ProcessImu)
        assert raw_data.shape == (3, 9, 100)
        assert len(processed_data) == 3
        processed_data = device.process(RealTimeProcessingMethod.ProcessImu, squared=True)
        assert raw_data.shape == (3, 9, 100)
        assert len(processed_data.shape) == 2


def test_marker_set():
    parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    model_path = parent_dir + "/examples/model/Wu_Shoulder_Model_mod_wt_wrapp.bioMod"
    marker_set = MarkerSet(nb_channels=16, name="my_marker_set", rate=100, system_rate=100)
    marker_set.data_window = 100
    assert marker_set.sample == 100 / 100
    assert marker_set.name == "my_marker_set"
    assert marker_set.nb_channels == 16

    np.random.seed(50)
    i = 0
    while i != 100:
        marker_set.new_data = np.random.random((3, 16, 1))
        marker_set.append_data(marker_set.new_data)
        i += 1

    raw_data = marker_set.raw_data
    kin_data, _ = marker_set.get_kinematics(model_path=model_path, method=InverseKinematicsMethods.BiorbdKalman)
    assert raw_data.shape == (3, 16, 100)
    assert kin_data.shape == (15, 1)



from .gui.plot import LivePlot, OfflinePlot

from .interfaces.pytrigno_interface import PytrignoClient
from .interfaces.vicon_interface import ViconClient
from .interfaces.generic_interface import GenericInterface
from .interfaces.tcp_interface import TcpClient
from .interfaces.param import Param, Device, MarkerSet


from .file_io.save_and_load import load, save

from .processing.data_processing import RealTimeProcessing, OfflineProcessing, GenericProcessing
from .processing.msk_functions import MskFunctions
from .processing.compute_mvc import ComputeMvc

from .streaming.client import Client, Message
from .streaming.server import Server
from .streaming.stream_data import StreamData

from .enums import *

__version__ = "2.0.0"



"""
This file contains a wrapper for the python Vicon SDK.
"""
from .param import *
from typing import Union
from .generic_interface import GenericInterface
from ..enums import InverseKinematicsMethods, InterfaceType

try:
    from vicon_dssdk import ViconDataStream as VDS
except ModuleNotFoundError:
    pass


class ViconClient(GenericInterface):
    """
    Class for interfacing with the Vicon system.
    """

    def __init__(self, system_rate: int, ip: str = "127.0.0.1", port: int = 801, init_now=True):
        """
        Initialize the ViconClient class.

        Parameters
        ----------
        system_rate: int
            Streaming rate of the nexus software.
        ip: str
            IP address of the nexus software.
        port: int
            Port of the nexus software.
        init_now: bool
            Whether to initialize the client now.
            Usefull if you want to pickle the interface as the Vicon SDK is not pickable (swig).
        """
        super(ViconClient, self).__init__(ip=ip, system_rate=system_rate, interface_type=InterfaceType.ViconClient)
        self.address = f"{ip}:{port}"

        self.vicon_client = None
        self.acquisition_rate = None
        self.system_rate = system_rate
        self.devices = []
        self.imu = []
        self.marker_sets = []
        self.is_frame = False
        self.is_initialized = False
        if init_now:
            self._init_client()

    def _init_client(self):
        """
        Initialize the Vicon client.
        """
        print(f"Connection to ViconDataStreamSDK at : {self.address} ...")
        self.vicon_client = VDS.Client()
        self.vicon_client.Connect(self.address)
        print("Connected to Vicon.")
        self.is_initialized = True

        # Enable several data types
        self.vicon_client.EnableSegmentData()
        self.vicon_client.EnableDeviceData()
        self.vicon_client.EnableMarkerData()
        self.vicon_client.EnableUnlabeledMarkerData()
        self.get_frame()
        if self.system_rate != self.vicon_client.GetFrameRate():
            raise ValueError(
                f"Vicon system rate ({self.vicon_client.GetFrameRate()}) does not match the system rate "
                f"({self.system_rate})."
            )

    def add_device(
        self,
        nb_channels: int,
        device_type: Union[DeviceType, str] = DeviceType.Emg,
        data_buffer_size: int = None,
        name: str = None,
        rate: float = 2000,
        device_range: tuple = None,
        processing_method: Union[RealTimeProcessingMethod, OfflineProcessingMethod] = None,
        **process_kwargs,
    ):
        """
        Add a device to the Vicon system.

        Parameters
        ----------
        nb_channels: int
            Number of channels of the device.
        device_type: Union[DeviceType, str]
            Type of the device.
        data_buffer_size: int
            Size of the buffer for the device.
        name: str
            Name of the device.
        rate: float
            Rate of the device.
        device_range: tuple
            Range of the device.
        processing_method : Union[RealTimeProcessingMethod, OfflineProcessingMethod]
            Method used to process the data.
        **process_kwargs
            Keyword arguments for the processing method.
        """
        device_tmp = self._add_device(
            nb_channels, device_type, name, rate, device_range, processing_method, **process_kwargs
        )
        device_tmp.interface = self.interface_type
        if self.vicon_client:
            device_tmp.infos = self.vicon_client.GetDeviceOutputDetails(name)
        else:
            device_tmp.infos = None
        device_tmp.data_windows = data_buffer_size
        self.devices.append(device_tmp)

    def add_marker_set(
        self,
        nb_markers: int,
        name: str = None,
        data_buffer_size: int = None,
        marker_names: Union[str, list] = None,
        rate: float = 100,
        unlabeled: bool = False,
        subject_name: str = None,
        kinematics_method: InverseKinematicsMethods = None,
        **kin_method_kwargs,
    ):
        """
        Add markers set to stream from the Vicon system.

        Parameters
        ----------
        nb_markers: int
            Number of markers.
        name: str
            Name of the markers set.
        data_buffer_size: int
            Size of the buffer for the markers set.
        marker_names: Union[list, str]
            List of markers names.
        rate: int
            Rate of the markers set.
        unlabeled: bool
            Whether the markers set is unlabeled.
        subject_name: str
            Name of the subject. If None, the subject will be the first one in Nexus.
        kinematics_method: InverseKinematicsMethods
            Method used to compute the kinematics.
        **kin_method_kwargs
            Keyword arguments for the kinematics method.
        """
        if len(self.marker_sets) != 0:
            raise ValueError("Only one marker set can be added for now.")

        markers_tmp = self._add_marker_set(
            nb_markers=nb_markers,
            name=name,
            marker_names=marker_names,
            rate=rate,
            unlabeled=unlabeled,
            kinematics_method=kinematics_method,
            **kin_method_kwargs,
        )
        if self.vicon_client:
            markers_tmp.subject_name = subject_name if subject_name else self.vicon_client.GetSubjectNames()[0]
            markers_tmp.marker_names = (
                self.vicon_client.GetMarkerNames(markers_tmp.subject_name) if not marker_names else marker_names
            )
            markers_tmp.marker_names = [name[0] for name in markers_tmp.marker_names]
            if markers_tmp.nb_channels != len(markers_tmp.marker_names):
                raise RuntimeError("Nb of marker not the same than markers on vicon.")
        else:
            markers_tmp.subject_name = subject_name
            markers_tmp.marker_names = marker_names
        markers_tmp.data_windows = data_buffer_size
        self.marker_sets.append(markers_tmp)

    @staticmethod
    def get_force_plate_data():
        raise NotImplementedError("Force plate streaming is not implemented yet.")

    def get_device_data(
        self, device_name: Union[str, list] = "all", channel_idx: Union[int, list] = (), get_frame: bool = True
    ):
        """
        Get the device data from Vicon.

        Parameters
        ----------
        device_name: str or list
            Name of the device or list of devices names.
        channel_idx: Union[int, str]
            Index of the channel to return.
        get_frame: bool
            Whether to get a new frame from the Vicon system.

        Returns
        -------
        device_data: list
            All asked device data.
        """
        if len(self.devices) == 0:
            raise ValueError("No device has been added to the Vicon system.")
        if not self.is_initialized:
            raise RuntimeError("Vicon client is not initialized.")
        if get_frame:
            self.get_frame()
        all_device_data = []
        if not isinstance(device_name, list):
            device_name = [device_name]
        if channel_idx and not isinstance(channel_idx, list):
            channel_idx = [channel_idx]
        device_data = []
        for d, device in enumerate(self.devices):
            if device_name[0] == "all" or device.name in device_name:
                device.new_data = np.zeros((device.nb_channels, device.sample))
                count = 0
                for output_name, channel_name, unit in device.infos:
                    data_tmp, _ = self.vicon_client.GetDeviceOutputValues(device.name, output_name, channel_name)
                    device.new_data[count, :] = data_tmp
                    device.channel_names.append(channel_name)
                    count += 1
                    if count == device.nb_channels:
                        break
                if channel_idx:
                    device_data = np.zeros((len(channel_idx), device.sample))
                    for idx in range(device.nb_channels):
                        if idx in channel_idx:
                            device_data[channel_idx.index(idx), :] = device.new_data[idx, :]
                device_data = device_data if channel_idx else device.new_data
                device.append_data(device.new_data)
                all_device_data.append(device_data)
        if len(all_device_data) == 1:
            return all_device_data[0]
        return all_device_data

    def get_marker_set_data(
        self, subject_name: Union[str, list] = None, marker_names: Union[str, list] = None, get_frame: bool = True
    ):
        """
        Get the markers data from Vicon.

        Parameters
        ----------
        subject_name: Union[str, list]
            Name of the subject. If None, the subject will be the first one in Nexus.
        marker_names: Union[str, list]
            List of markers names.
        get_frame: bool
            Whether to get a new frame or not.

        Returns
        -------
        markers_data: list
            All asked markers data.
        """
        if len(self.marker_sets) == 0:
            raise ValueError("No marker set has been added to the Vicon system.")
        if not self.is_initialized:
            raise RuntimeError("Vicon client is not initialized.")
        if get_frame:
            self.get_frame()
        if subject_name and isinstance(subject_name, list):
            subject_name = [subject_name]
        if marker_names and isinstance(marker_names, list):
            marker_names = [marker_names]
        occluded = []
        all_markers_data = []
        all_occluded_data = []
        if subject_name:
            marker_sets = [None] * len(subject_name)
            for s, marker_set in enumerate(self.marker_sets):
                if marker_set.subject_name in subject_name:
                    marker_sets[subject_name.index(marker_set.subject_name)] = marker_set
            if marker_sets == [None]:
                raise RuntimeError("No subject of this name.")
        else:
            marker_sets = self.marker_sets

        for markers in marker_sets:
            markers.new_data = np.zeros((3, len(markers.marker_names), markers.sample))
            count = 0
            for m, marker_name in enumerate(markers.marker_names):
                markers_data_tmp, occluded_tmp = self.vicon_client.GetMarkerGlobalTranslation(
                    markers.subject_name, marker_name
                )
                markers.new_data[:, count, :] = np.array(markers_data_tmp)[:, np.newaxis]
                occluded.append(occluded_tmp)
            if marker_names:
                markers_data = np.zeros((3, len(marker_names), markers.sample))
                for n, name in enumerate(markers.marker_names):
                    if name in marker_names:
                        markers_data[:, marker_names.index(name), :] = markers.new_data[:, n, :]
                all_markers_data.append(markers_data)
            else:
                all_markers_data.append(markers.new_data)

            markers.append_data(markers.new_data)
            all_occluded_data.append(occluded)
        if len(all_markers_data) == 1:
            return all_markers_data[0], all_occluded_data[0]
        return all_markers_data, all_occluded_data

    def init_client(self):
        """
        Initialize the Vicon client if it is not already initialized.
        This function has to be called before get frame from interface.
        """
        if self.is_initialized:
            raise RuntimeError("Vicon client is already initialized.")
        else:
            self._init_client()
            for d, device in enumerate(self.devices):
                if not device.infos:
                    device.infos = self.vicon_client.GetDeviceOutputDetails(device.name)
            for m, marker_set in enumerate(self.marker_sets):
                if not marker_set.markers_names:
                    marker_set.markers_names = self.vicon_client.GetMarkerNames(marker_set.subject_name)
                if not marker_set.subject_name:
                    marker_set.subject_name = self.vicon_client.GetSubjectNames()[0]

    def get_latency(self) -> float:
        """
        Get the latency between the Vicon system and the Vicon SDK.

        Returns
        -------
        latency: float
            Latency between the Vicon system and the Vicon SDK.
        """
        if not self.is_initialized:
            raise RuntimeError("Vicon client is not initialized.")
        return self.vicon_client.GetLatencyTotal()

    def get_frame(self) -> bool:
        """
        Get a new frame from the Vicon system.

        Returns
        -------
        bool
            True if there is a frame, False otherwise.
        """
        if not self.is_initialized:
            raise RuntimeError("Vicon client is not initialized.")
        self.is_frame = self.vicon_client.GetFrame()
        while self.is_frame is not True:
            self.is_frame = self.vicon_client.GetFrame()
        return self.is_frame

    def get_frame_number(self) -> int:
        """
        Get the last frame number.

        Returns
        -------
        frame_number: int
            Last frame number.
        """
        if not self.is_initialized:
            raise RuntimeError("Vicon client is not initialized.")
        return self.vicon_client.GetFrameNumber()

    def get_kinematics_from_markers(
        self,
        marker_set_name: str,
        model_path: str = None,
        method: Union[InverseKinematicsMethods, str] = InverseKinematicsMethods.BiorbdLeastSquare,
        custom_func: callable = None,
        **kwargs,
    ):
        """
        Get the kinematics from markers.

        Parameters
        ----------
        marker_set_name: str
            name of the markerset.
        model_path: str
            biorbd model of the kinematics.
        method: str
            Method to use to get the kinematics. Can be "kalman" or "custom".
        custom_func: function
            Custom function to get the kinematics.

        Returns
        -------
        kinematics: list
            List of kinematics.
        """
        marker_set_idx = [i for i, m in enumerate(self.marker_sets) if m.name == marker_set_name][0]
        return self.marker_sets[marker_set_idx].get_kinematics(model_path, method, custom_func=custom_func, **kwargs)



"""
This example shows how to apply an inverse kinematics method to a biorbd model and a set of markers.
To use these methods, you will need to use and install the biorbd(https://github.com/pyomeca/biorbd) and bioviz (https://github.com/pyomeca/bioviz) libraries. More information is available on the projects GitHub page. This example uses a marker data stream from the Vicon Nexus interface.
If you want to try this example offline, you can use the provided custom interface, called 'MyInterface', which you can use as a standard interface.
For more information on how to do this, please refer to the marker_streaming.py example. Here, the processing method used is the inverse kinematics method of the biorbd library. The implemented methods are available in the class InverseKinematicsMethods. This method takes the biorbd model as an argument in the initialization.
The processing method will return the angle and velocity of the model joint. You can display the result using the skeleton plot which uses the bioviz library. In the skeleton plot initialization, you can specify the arguments belonging to the bioviz.Viz function. For more information about the arguments, please refer to the bioviz documentation.
Be aware that the skeleton plot may take some time and slow down the loop. If you want to display the data in real time, consider using a lightweight *.vtp file inside the model.
"""
import time

import numpy as np

from biosiglive import ViconClient, InverseKinematicsMethods, LivePlot, PlotType, InterfaceType
from custom_interface import MyInterface


if __name__ == "__main__":
    interface = InterfaceType.Custom
    if interface == InterfaceType.Custom:
        interface = MyInterface(system_rate=100, data_path="abd.bio")
    else:
        interface = ViconClient(ip="127.0.0.1", system_rate=100)

    interface.add_marker_set(
        nb_markers=16,
        data_buffer_size=100,
        marker_data_file_key="markers",
        name="markers",
        rate=100,
        kinematics_method=InverseKinematicsMethods.BiorbdKalman,
        model_path="model/Wu_Shoulder_Model_mod_wt_wrapp.bioMod",
        unit="mm",
    )

    marker_plot = LivePlot(name="markers", plot_type=PlotType.Skeleton)
    marker_plot.init(model_path="model/Wu_Shoulder_Model_mod_wt_wrapp.bioMod", show_floor=False, show_muscles=False)
    while True:
        tic = time.time()
        Q, Qdot = interface.get_kinematics_from_markers(marker_set_name="markers", get_markers_data=True)
        marker_plot.update(Q)
        if interface == InterfaceType.Custom:
            time.sleep((1 / 100) - (time.time() - tic))



"""
This file contains a wrapper to use a tcp client more easily.
"""
from ..streaming.client import Client, Message
from .generic_interface import GenericInterface
from ..enums import (
    DeviceType,
    InterfaceType,
    RealTimeProcessingMethod,
    OfflineProcessingMethod,
    InverseKinematicsMethods,
)
from typing import Union


class TcpClient(GenericInterface):
    """
    Class for interfacing with the client.100
    """

    def __init__(self, ip: str = "127.0.0.1", port: int = 801, client_type: str = "TCP", read_frequency: int = 100):
        """
        Initialize the client.

        Parameters
        ----------
        ip: str
            IP address of the server.
        port: int
            Port of the server.
        client_type: str
            Type of the server.
        read_frequency: int
            Frequency of the reading of the data.
        """
        super(TcpClient, self).__init__(ip, interface_type=InterfaceType.TcpClient)
        self.devices = []
        self.imu = []
        self.marker_sets = []
        self.read_frequency = read_frequency
        self.ip = ip
        self.port = port
        self.device_cmd_names = []
        self.marker_cmd_names = []
        self.last_server_data = None
        self.client = Client(server_ip=ip, port=port, client_type=client_type)

    def add_device(
        self,
        nb_channels: int,
        device_type: Union[DeviceType, str] = DeviceType.Emg,
        command_name: Union[str, list] = "",
        name: str = None,
        rate: float = 2000,
        device_range: tuple = None,
        processing_method: Union[RealTimeProcessingMethod, OfflineProcessingMethod] = None,
        **process_kwargs,
    ):
        """
        Add a device to the client.

        Parameters
        ----------
        nb_channels: int
            Number of channels of the device.
        device_type: Union[DeviceType, str]
            Type of the device. (emg, imu, etc.)
        command_name: Union[str, list]
            Name of the command to send to the server.
        name: str
            Name of the device.
        rate: float
            Frequency of the device.
        device_range: tuple
            Range of the device.
        processing_method: Union[RealTimeProcessingMethod, OfflineProcessingMethod]
            Method to use to process the data.
        process_kwargs: dict
            Keyword arguments for the processing method.
        """
        device_tmp = self._add_device(
            nb_channels, device_type, name, rate, device_range, processing_method, **process_kwargs
        )
        device_tmp.interface = self.interface_type
        self.devices.append(device_tmp)
        self.device_cmd_names.append(command_name)

    def add_marker_set(
        self,
        nb_markers: int,
        name: str = None,
        command_name: Union[str, list] = None,
        marker_names: Union[str, list] = None,
        rate: float = 100,
        unlabeled: bool = False,
        subject_name: str = None,
        kinematics_method: InverseKinematicsMethods = None,
        **kin_method_kwargs,
    ):
        """
        Add markers set to stream from the tcp interface.

        Parameters
        ----------
        nb_markers: int
            Number of markers.
        name: str
            Name of the markers set.
        command_name: Union[str, list]
            Name of the command to send to the server.
        marker_names: Union[list, str]
            List of markers names.
        rate: int
            Rate of the markers set.
        unlabeled: bool
            Whether the markers set is unlabeled.
        subject_name: str
            Name of the subject. If None, the subject will be the first one in Nexus.
        kinematics_method: InverseKinematicsMethods
            Method used to compute the kinematics.
        **kin_method_kwargs
            Keyword arguments for the kinematics method.
        """
        if len(self.marker_sets) != 0:
            raise ValueError("Only one marker set can be added for now.")

        markers_tmp = self._add_marker_set(
            nb_markers=nb_markers,
            name=name,
            marker_names=marker_names,
            rate=rate,
            unlabeled=unlabeled,
            kinematics_method=kinematics_method,
            **kin_method_kwargs,
        )
        self.marker_sets.append(markers_tmp)
        self.marker_cmd_names.append(command_name)

    def get_data_from_server(
        self, command: Union[str, list] = "all", nb_frame_to_get: int = None, down_sampling: dict = None
    ) -> dict:
        """
        Get the data from the server.

        Parameters
        ----------
        command: Union[str, list]
            Command to send to the server.
        nb_frame_to_get: int
            Number of frame to get.
        down_sampling: dict
            Down sampling parameters. Keys are the names of the devices and values are the down sampling rate.

        Returns
        -------
        data: dict
            Data from the server.
        """
        self.last_server_data = self.client.get_data(
            message=Message(command=command, nb_frame_to_get=nb_frame_to_get, down_sampling=down_sampling)
        )
        return self.last_server_data

    def get_device_data(
        self,
        device_name: Union[list, str] = "all",
        nb_frame_to_get: int = None,
        down_sampling: dict = None,
        get_from_server: bool = True,
    ):
        """
        Get the data from a device.

        Parameters
        ----------
        device_name:  Union[list, str]
            Name of the device. All for all the devices.
        nb_frame_to_get: int
            Number of frame to get.
        down_sampling: dict
            Down sampling parameters. Keys are the names of the devices and values are the down sampling rate.
        get_from_server: bool
            Whether to get the data from the server or not.

        Returns
        -------
        data: list
            The data asked from the server.
        """
        if get_from_server:
            command = self._prepare_cmd(device_name, True)
            data = self.get_data_from_server(
                command=command, nb_frame_to_get=nb_frame_to_get, down_sampling=down_sampling
            )
        else:
            if nb_frame_to_get or down_sampling:
                raise ValueError("nb_frame_to_get and down_sampling can only be used if get_from_server is True.")
            if not self.last_server_data:
                raise ValueError("No data received from the server yet.")
            data = self.last_server_data

        all_data = []
        for d, device in enumerate(self.devices):
            if device_name == "all" or device.name in device_name:
                if self.device_cmd_names[d] not in data.keys():
                    raise ValueError(f"No data for device '{device.name}'. Check the command name.")
                device.new_data = data[self.device_cmd_names[d]][d]
                device.append_data(device.new_data)
                all_data.append(device.new_data)
        if len(all_data) == 1:
            return all_data[0]
        return all_data

    def get_marker_set_data(
        self,
        marker_set_name: Union[str, list] = "all",
        nb_frame_to_get: int = None,
        down_sampling: dict = None,
        get_from_server: bool = True,
    ):
        """
        Get the data from the markers.

        Parameters
        ----------
        marker_set_name:  Union[list, str]
            Name of the markers set. all for all the markers sets.
        nb_frame_to_get: int
            Number of frame to get.
        down_sampling: dict
            Down sampling parameters. Keys are the names of the devices and values are the down sampling rate.
        get_from_server: bool
            Whether to get the data from the server or not.

        Returns
        -------
        data: list
            The data asked from the server.
        """
        if get_from_server:
            command = self._prepare_cmd(marker_set_name, False)
            data = self.get_data_from_server(
                command=command, nb_frame_to_get=nb_frame_to_get, down_sampling=down_sampling
            )
        else:
            if not self.last_server_data:
                raise ValueError("No data received from the server yet.")
            data = self.last_server_data

        all_data = []
        for m, marker_set in enumerate(self.marker_sets):
            if marker_set_name == "all" or marker_set.name in marker_set_name:
                if self.marker_cmd_names[m] not in data.keys():
                    raise ValueError(f"No data for marker set '{marker_set.name}'. Check the command name.")
                marker_set.new_data = data[self.marker_cmd_names[m]][m]
                marker_set.append_data(marker_set.new_data)
                all_data.append(marker_set.new_data)
        if len(all_data) == 1:
            return all_data[0]
        return all_data

    def _prepare_cmd(self, name: Union[str, list], device: bool = True):
        """
        Prepare the command to send to the server.

        Parameters
        ----------
        name: Union[str, list]
            Name of the device or marker set.

        Returns
        -------
        cmd: str
            Command to send to the server.
        """
        if not isinstance(name, list):
            name = [name]

        command = self.device_cmd_names if device else self.marker_cmd_names
        stream_param = self.devices if device else self.marker_sets
        commands = []
        if name != "all":
            for s, param in enumerate(stream_param):
                if param.name == name[s]:
                    commands.append(command[s])
        return commands



"""
This file is part of biosiglive. it is an example to see how to use biosiglive to compute the maximal voluntary
 contraction from EMG signals.
"""

from time import strftime
from ..interfaces.vicon_interface import ViconClient
from ..interfaces.pytrigno_interface import PytrignoClient
from ..interfaces.tcp_interface import TcpClient
from ..interfaces.generic_interface import GenericInterface
from .data_processing import OfflineProcessing
from ..gui.plot import LivePlot, OfflinePlot
from ..enums import InterfaceType
from ..file_io.save_and_load import save, load
from pathlib import Path
from time import time, sleep
import os
import numpy as np
from typing import Union


# TODO: Add the possibility to send a temporary file to append the data to it in case of former dysfunction.
class ComputeMvc:
    def __init__(
        self,
        interface_type: Union[str, InterfaceType] = InterfaceType.PytrignoClient,
        interface_ip: str = "127.0.0.1",
        interface_port: int = 801,  # only for vicon
        output_file: str = None,
        muscle_names: list = None,
        emg_frequency: float = 2000,
        acquisition_rate: int = 100,
        mvc_windows: int = 2000,
        range_muscle: Union[tuple, int] = None,
        custom_interface: GenericInterface = None,
    ):
        """
        Initialize the class.

        Parameters
        ----------
        interface_type : Union[str, InterfaceType]
            The mode of the stream (describe in the InterfaceType enum).
        interface_ip : str
            The ip of the interface.
        interface_port : int
            The port of the interface. Only needed for the Vicon interface.
        output_file : str
            The path of the output file.
        muscle_names : list
            The list of the muscle names.
        emg_frequency : float
            The frequency of the device.
        acquisition_rate : float
            The acquisition rate of the acquisition.
        mvc_windows : int
            size of the window to compute the mvc.
        range_muscle : tuple
            The range of the muscle to compute the mvc.
        custom_interface: classmethod
            The custom interface to use
        """
        if Path(output_file).suffix != ".bio":
            if Path(output_file).suffix == "":
                output_file += ".bio"
            else:
                raise ValueError("The file must be a .bio file.")

        if isinstance(interface_type, str):
            if interface_type not in [t.value for t in InterfaceType]:
                raise ValueError("The type of the interface is not valid.")
            interface_type = InterfaceType(interface_type)
        self.interface_type = interface_type
        self.range_muscle = range_muscle if range_muscle else (0, 16)
        self.nb_muscles = self.range_muscle[1] - self.range_muscle[0]
        if muscle_names:
            self.muscle_names = muscle_names
        else:
            self.muscle_names = []
            for i in range(self.range_muscle[0], self.range_muscle[1]):
                self.muscle_names.append(f"Muscle_{i}")
        if self.nb_muscles != len(self.muscle_names):
            raise RuntimeError("number of muscle must have the same length than names.")
        self.frequency = emg_frequency
        self.acquisition_rate = acquisition_rate
        self.mvc_windows = mvc_windows

        current_time = strftime("%Y%m%d-%H%M")
        self.output_file = f"MVC_{current_time}.bio" if not output_file else output_file

        self.device_host = None
        self.interface_port = interface_port
        self.interface_ip = interface_ip
        self.device_name = None
        self.show_data = False
        self.is_processing_method = False
        self.try_number = 0
        self.emg_plot = None

        self.emg_processing = None
        self.moving_average, self.low_pass, self.custom = None, None, None

        self.try_name = ""
        self.try_list = []
        self.emg_interface = None

        if self.interface_type == InterfaceType.PytrignoClient:
            self.emg_interface = PytrignoClient(system_rate=self.acquisition_rate, ip=self.interface_ip)
        elif self.interface_type == InterfaceType.ViconClient:
            self.emg_interface = ViconClient(
                system_rate=self.acquisition_rate, ip=self.interface_ip, port=self.interface_port
            )
        elif self.interface_type == InterfaceType.TcpClient:
            self.emg_interface = TcpClient(
                read_frequency=self.acquisition_rate, ip=self.interface_ip, port=self.interface_port
            )
        elif self.interface_type == InterfaceType.Custom:
            self.emg_interface = custom_interface

        if self.interface_type != InterfaceType.Custom:
            self.emg_interface.add_device(nb_channels=self.nb_muscles, name="EMG", device_range=self.range_muscle)
        else:
            if len(self.emg_interface.devices) == 0:
                raise RuntimeError("Please add a device in the custom interface.")

    def set_processing_method(
        self,
        moving_average: bool = True,
        low_pass: bool = False,
        custom: bool = False,
        custom_function: callable = None,
        bandpass_frequency: tuple = (10, 425),
        lowpass_frequency: float = 5,
        lowpass_order: int = 4,
        butterworth_order: int = 4,
        ma_window: int = 200,
    ):
        """
        Set the emg processing method. This method allow to customize the processing of the emg signal,
        so it need to be called before the start of the acquisition.

        Parameters
        ----------
        moving_average : bool
            If True, the emg data will be processed with a moving average.
        low_pass : bool
            If True, the emg data will be processed with a low pass filter.
        custom : bool
            If True, the emg data will be processed with a custom function.
        custom_function : callable
            The custom function. Input : raw data, device frequency Output : processed data.
        bandpass_frequency : tuple
            The frequency of the bandpass filter.
        lowpass_frequency : float
            The frequency of the low pass filter.
        lowpass_order : int
            The order of the low pass filter.
        butterworth_order : int
            The order of the butterworth filter.
        ma_window : int
            The size of the moving average window.
        """

        self.moving_average = moving_average
        self.low_pass = low_pass
        self.custom = custom
        if [moving_average, custom, low_pass].count(True) > 1:
            raise ValueError("Only one processing method can be selected")
        if custom and not custom_function:
            raise ValueError("custom_function must be defined")
        if custom:
            self.emg_processing = custom_function
        else:
            emg_processing = OfflineProcessing()
            emg_processing.bpf_lcut = bandpass_frequency[0]
            emg_processing.bpf_hcut = bandpass_frequency[1]
            emg_processing.lpf_lcut = lowpass_frequency
            emg_processing.lp_butter_order = lowpass_order
            emg_processing.bp_butter_order = butterworth_order
            emg_processing.ma_win = ma_window
            self.emg_processing = emg_processing.process_emg
        self.is_processing_method = True

    def run(self, show_data: bool = False) -> list:
        """
        Run the MVC program.

        Parameters
        ----------
        show_data: bool
            If True, the data will be displayed in live in a separate plot.

        Returns
        -------
        list
            The list of the MVC value for each muscle.
        """
        self.show_data = show_data
        self.try_number = 0
        while True:
            if show_data:
                self.emg_plot = LivePlot(nb_subplots=self.nb_muscles, channel_names=self.muscle_names)
                self.emg_plot.init(plot_windows=int(self.frequency * 2))
            nb_frame, var, duration = self._init_trial()
            trial_emg = self._mvc_trial(duration, nb_frame, var)
            processed_emg, raw_emg = self._process_emg(trial_emg, save_tmp=True)
            self._plot_trial(raw_emg, processed_emg)

            task = input(
                "Press 'c' to do an other MVC trial," " 'r' to do again the MVC trial or 'q' then enter to quit.\n"
            )

            while task != "c" and task != "r" and task != "q":
                print(f"Invalid entry ({task}). Please press 'c', 'r', or 'q' (in lowercase).")
                task = input(
                    "Press 'c' to do an other MVC trial," " 'r' to do again the MVC trial or 'q' then enter to quit.\n"
                )

            if task == "c":
                pass

            elif task == "r":
                self.try_number -= 1
                mat_content = load("_MVC_tmp.bio")
                mat_content.pop(f"{self.try_name}_processed", None)
                mat_content.pop(f"{self.try_name}_raw", None)
                save(mat_content, "_MVC_tmp.bio")
                self.try_list = self.try_list[:-1]

            elif task == "q":
                mvc_list = self._save_trial()
                break
        return mvc_list

    def _init_trial(self):
        """
        Initialize the trial.

        Returns
        -------
        nb_frame : int
            The number of frames of the trial.
        var : float
            The current iteration.
        duration : float
            The duration of the trial.
        """

        try_name = input("Please enter a name of your trial (string) then press enter or press enter.\n")
        while try_name in self.try_list:
            try_name = input("This name is already used. Please chose and other name.\n")

        if try_name == "":
            self.try_name = f"MVC_{self.try_number}"
        else:
            self.try_name = f"{try_name}"
        self.try_number += 1

        self.try_list.append(self.try_name)
        t = input(
            f"Ready to start trial: {self.try_name}, with muscles :{self.muscle_names}. "
            f"Press enter to begin your MVC. or enter a number of seconds."
        )
        nb_frame = 0
        try:
            float(t)
            iter = float(t) * self.acquisition_rate
            var = iter
            duration = True
        except ValueError:
            var = -1
            duration = False
        return nb_frame, var, duration

    def _mvc_trial(self, duration: float, nb_frame: int, var: float):
        """
        Run the MVC trial.

        Parameters
        ----------
        duration : float
            The duration of the trial.
        nb_frame : int
            The number of frames of the trial.
        var : float
            The current iteration.

        Returns
        -------
        trial_emg : numpy.ndarray
            The EMG data of the trial.
        """
        data = None
        while True:
            try:
                if nb_frame == 0:
                    print("Trial is running please press 'Ctrl+C' when trial is ended (it will not end the program).")

                data_tmp = self.emg_interface.get_device_data()
                tic = time()

                data = data_tmp if nb_frame == 0 else np.append(data, data_tmp, axis=1)
                if self.show_data:
                    self.emg_plot.update(data_tmp)
                nb_frame += 1

                time_to_sleep = (1 / self.acquisition_rate) - (time() - tic)

                if time_to_sleep > 0:
                    sleep(time_to_sleep)
                else:
                    print(f"Delay of {abs(time_to_sleep)}.")

                if duration:
                    if nb_frame == var:
                        return data

            except KeyboardInterrupt:
                if self.show_data is True:
                    self.emg_plot.disconnect()
                return data

    def _plot_trial(self, raw_data: np.ndarray = None, processed_data: np.ndarray = None):
        """
        Plot the trial.

        Parameters
        ----------
        raw_data : numpy.ndarray
            The raw EMG data of the trial.
        processed_data : numpy.ndarray
                The processed EMG data of the trial.
        """
        data = raw_data
        legend = ["Raw"]
        nb_column = 4 if raw_data.shape[0] > 4 else raw_data.shape[0]
        n_p = 0
        plot_comm = "y"
        print(f"Trial {self.try_name} terminated. ")
        while plot_comm != "n":
            if n_p != 0:
                plot_comm = input(f"Would you like to plot again ? 'y'/'n'")

            if plot_comm == "y":
                plot = input(
                    f"Press 'pr' to plot your raw trial,"
                    f" 'p' to plot your processed trial, 'b' to plot both or 'c' to continue,"
                    f" then press enter."
                )
                while plot != "p" and plot != "pr" and plot != "c" and plot != "b":
                    print(f"Invalid entry ({plot}). Please press 'p', 'pr', 'b',  or 'c' (in lowercase).")
                    plot = input(
                        f"Press 'pr' to plot your raw trial,"
                        f"'p' to plot your processed trial or 'c' to continue then press enter."
                    )

                if plot != "c":
                    if plot == "p":
                        data = processed_data
                        legend = ["Processed"]
                    elif plot == "b":
                        data = [raw_data, processed_data]
                        legend = ["Raw data", "Processed"]
                    legend = legend * raw_data.shape[0]
                    x = np.linspace(0, raw_data.shape[1] / self.frequency, raw_data.shape[1])
                    print("Close the plot windows to continue.")
                    OfflinePlot().multi_plot(
                        data,
                        nb_column=nb_column,
                        y_label="Activation level (v)",
                        x_label="Time (s)",
                        legend=legend,
                        subplot_title=self.muscle_names,
                        figure_name=self.try_name,
                        x=x,
                    )
                else:
                    pass
                n_p += 1

    def _process_emg(self, data, save_tmp=True):
        """
        Process the EMG data.

        Parameters
        ----------
        data : numpy.ndarray
            The raw EMG data of the trial.
        save_tmp : bool
            If True, the processed data is saved in a temporary file.

        Returns
        -------
        numpy.ndarray
            The processed EMG data of the trial.
        """
        if not self.is_processing_method:
            self.set_processing_method()
        emg_processed = self.emg_processing(
            data, data_rate=self.frequency, moving_average=self.moving_average, low_pass_filter=self.low_pass
        )
        file_name = "_MVC_tmp.bio"
        # save tmp_file
        if save_tmp:
            if os.path.isfile(file_name):
                mat = load(file_name)
                mat[f"{self.try_name}_processed"] = emg_processed
                mat[f"{self.try_name}_raw"] = data
                data_to_save = mat
            else:
                data_to_save = {f"{self.try_name}_processed": emg_processed, f"{self.try_name}_raw": data}
            save(data_to_save, file_name)
        return emg_processed, data

    def _save_trial(self) -> list:
        """
        Save and end the trial.

        Returns
        -------
        list
            The list of the mvc for each muscle.
        """

        print("Concatenate data for all trials.")
        mat_content = load("_MVC_tmp.bio")
        data_final = []
        for i in range(len(self.try_list)):
            if i == 0:
                data_final = mat_content[f"{self.try_list[i]}_processed"]
            else:
                data_final_tmp = mat_content[f"{self.try_list[i]}_processed"]
                data_final = np.append(data_final, data_final_tmp, axis=1)

        save = input("Press 's' to save your data, other key to just return a list of MVC.\n")
        if save != "s":
            save = input("Data will not be saved. " "If you want to save press 's', if not, press enter.\n")

        print("Please wait during data processing (it could take some time)...")
        emg_processed, data_raw = self._process_emg(data_final, save_tmp=False)
        mvc_trials = emg_processed
        save = True if save == "s" else False
        mvc = OfflineProcessing.compute_mvc(
            self.nb_muscles, mvc_trials, self.mvc_windows, "_MVC_tmp.bio", self.output_file, save
        )
        return mvc



"""
This example shows how to use the computeMvc class to calculate the MVC of a online trial set.
You can use an implemented interface and specify it to the computeMvc class by specifying what type of pre-implemented interface you are using.
Once started, the computeMVC class will start in the terminal with questions and you will have to answer them.
Note that if you use this function outside the terminal (pycharm for example) you will have to emulate the terminal in the command window (for pycharm: Run -> Edit Configurations -> Emulate terminal in output console).
First the program will ask you the name of the test you are doing, then the number of seconds you want for the test, it will end at the end (if not specified, it will continue until you press ctrl+c). At the end of the trial, the data will be processed. If you want to change the processing method, you can do so via the set_processing_method before calling the run() method.
You can plot the data (raw, processed or both) at the end of the processing.
Then you can continue with other trials and do the same commands until you decide to stop the program at the end of a trial. In case of malfunction, a temporary file is saved after each trial with all previously recorded data (raw and processed). At the end of all trials, the MVC will be calculated by sorting the data and taking the highest values (for the specified MVC_window). You can then save the data to a file or simply terminate the program which will return a list of MVCs of the length of the number of muscles.
"""
from biosiglive import ComputeMvc, InterfaceType
from custom_interface import MyInterface

if __name__ == "__main__":
    custom = True
    custom_interface = MyInterface(100, "abd.bio")
    custom_interface.add_device(5, "emg", "my_emg", device_data_file_key="emg")
    interface_type = InterfaceType.Custom if custom else InterfaceType.ViconClient
    list_mvc = ComputeMvc(
        interface_type=interface_type,
        interface_ip="127.0.0.1",
        range_muscle=(0, 5),
        output_file="mvc.bio",
        custom_interface=custom_interface,
    ).run()
    print(list_mvc)



import numpy as np


def dic_merger(dic_to_merge, new_dic=None):
    """Merge two dictionaries.

    Parameters
    ----------
    dic_to_merge : dict
        Dictionary to merge.
    new_dic : dict
        Dictionary to merge with.

    Returns
    -------
    dict
        Merged dictionary.
    """

    if not new_dic:
        new_dic = dic_to_merge
    else:
        for key in dic_to_merge.keys():
            if isinstance(dic_to_merge[key], dict):
                new_dic[key] = dic_merger(dic_to_merge[key], new_dic[key])
            elif isinstance(dic_to_merge[key], list):
                new_dic[key] = dic_to_merge[key] + new_dic[key]
            elif isinstance(dic_to_merge[key], np.ndarray):
                new_dic[key] = np.append(dic_to_merge[key], new_dic[key], axis=0)
            elif isinstance(dic_to_merge[key], int):
                if isinstance(new_dic[key], int):
                    new_dic[key] = [new_dic[key]]
                new_dic[key] = [new_dic[key]] + [dic_to_merge[key]]
    return new_dic



"""
This example shows how to use the biosiglive library to get the cadence of a treadmill.
First, an interface is created (here Vicon) and then a device (here a treadmill) is added to the interface. If you want to try this example offline, you can use the provided custom interface named "MyInterface" and you can use it as a standard interface.
The data can be plotted in real time at each loop, please see the live_plot.py example.

"""
from biosiglive.interfaces.vicon_interface import ViconClient
from time import time, sleep
from custom_interface import MyInterface
from biosiglive import RealTimeProcessingMethod, InterfaceType, DeviceType


if __name__ == "__main__":
    interface = None
    interface_type = InterfaceType.Custom
    if interface_type == InterfaceType.Custom:
        interface = MyInterface(system_rate=100, data_path="walk.bio")
    elif interface_type == InterfaceType.ViconClient:
        interface = ViconClient(system_rate=100)
    nb_second = 30
    interface.add_device(
        9,
        name="Treadmill",
        device_type=DeviceType.Generic,
        rate=1000,
        processing_method=RealTimeProcessingMethod.GetPeaks,
        data_buffer_size=1000 * nb_second,
        processing_window=1000 * nb_second,
        device_data_file_key="treadmill",
        threshold=0.6,
        min_peaks_interval=1300,
    )
    print_every = 10  # seconds
    count = 0
    tic_bis = time()
    while True:
        data = interface.get_device_data(device_name="Treadmill")
        force_z_tmp = data[[2, 8], :]
        cadence, force_z_process = interface.get_device("Treadmill").process()
        if count == print_every * interface.devices[-1].system_rate:
            print(f"Mean cadence for the last {nb_second} s is :{cadence}")
            tic_bis = time()
            count = 0
        count += 1
        if interface_type == InterfaceType.Custom:
            sleep(1 / 100)



"""
This file contains a wrapper for the socket server to send data to the server.
"""
import socket
import numpy as np
import struct

try:
    from pythonosc.udp_client import SimpleUDPClient
except ModuleNotFoundError:
    pass
import pickle


class Connection:
    """
    This class is used to connect to the biosiglive server.
    """

    def __init__(self, ip: str = "127.0.0.1", port: int = 50000):
        """
        Initialize the connection.

        Parameters
        ----------
        ip : str
            The ip address of the server.
        port : int
            The port of the server.
        """
        self.ip = ip
        self.ports = port
        self.message_queues = None
        self.buff_size = 100000
        self.acquisition_rate = 100

    @staticmethod
    def _prepare_data(command: list, data: dict, down_sampling: dict = None, nb_frames_to_get: int = None):
        """
        Prepare the data to send.

        Parameters
        ----------
        command : dict
            The command received from the client.
        data : dict
            The data to prepared.
        down_sampling : dict
            The down sampling to apply to the data.
        nb_frames_to_get : int
            The number of frames to get.

        Returns
        -------
        prepared data : dict
            The data prepared to be sent.

        """
        data_tmp = {}
        if command == "all":
            command = list(data.keys())
        for key in command:
            if key in data.keys():
                data_tmp[key] = []
                for d in data[key]:
                    if down_sampling and key in down_sampling.keys():
                        if not isinstance(down_sampling[key], int):
                            raise ValueError("The down sampling must be an integer.")
                        d = d[..., :: down_sampling[key]]
                    if isinstance(d, np.ndarray):
                        if nb_frames_to_get:
                            data_tmp[key].append(d[..., -nb_frames_to_get:])
                        else:
                            data_tmp[key].append(d)
                    else:
                        data_tmp[key].append(d)
            else:
                raise ValueError(f"The asked data '{key}' is not in the data dictionary.")
            for key in data_tmp.keys():
                if len(data_tmp[key]) == 1:
                    data_tmp[key] = data_tmp[key][0]
        return data_tmp


class Server(Connection):
    """
    Class to create a server.
    """

    def __init__(self, ip: str = "127.0.0.1", port: int = 50000, server_type: str = "TCP"):
        """
        Initialize the server.

        Parameters
        ----------
        ip : str
            The ip of the server.
        port : int
            The port of the server.
        server_type : str
            The type of the server.
        """
        self.ip = ip
        self.port = port
        self.server_type = server_type
        self.server = None
        self.inputs = None
        self.outputs = None
        self.buff_size = 100000
        super().__init__(ip=ip, port=port)

    def start(self):
        """
        Start the server.
        """
        if self.server_type == "TCP":
            self.server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        elif self.server_type == "UDP":
            raise RuntimeError(f"UDP server not implemented yet.")
            # self.servers.append(socket.socket(socket.AF_INET, socket.SOCK_DGRAM))
        else:
            raise RuntimeError(f"Invalid type of connexion ({self.server_type}). Type must be 'TCP' or 'UDP'.")
        try:
            self.server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self.server.bind((self.ip, self.port))
            if self.server_type != "UDP":
                self.server.listen(10)
                self.inputs = [self.server]
                self.outputs = []
                self.message_queues = {}
        except ConnectionError:
            raise RuntimeError("Unknown error. Server is not listening.")

    def client_listening(self):
        """
        Waiting for the client connection.
        """
        connection, ad = self.server.accept()
        message = pickle.loads(connection.recv(self.buff_size))  # Received message
        return connection, message

    def send_data(self, data: dict, connection: socket.socket, message: dict = None):
        """
        Send the data to the client.

        Parameters
        ----------
        data : dict
            The data to send.
        connection : socket.socket
            The connection to send the data to.
        message : dict
            The message received from the client.
        """
        if message:
            if (
                "command" in message.keys()
                and "down_sampling" in message.keys()
                and "nb_frames_to_get" in message.keys()
            ):
                data = self._prepare_data(
                    message["command"], data, message["down_sampling"], message["nb_frames_to_get"]
                )
            else:
                raise ValueError(
                    "The message should be a dictionary created from the Message class or contains the key"
                    " : 'command', down_sampling, nb_frames_to_get."
                )
        encoded_data = pickle.dumps(data)
        encoded_data = struct.pack(">I", len(encoded_data)) + encoded_data
        try:
            connection.sendall(encoded_data)
        except ConnectionError:
            raise RuntimeError("Unknown error. Data not sent.")


class OscClient(Connection):
    """
    Class to create an OSC client.
    """

    def __init__(self, ip: str = "127.0.0.1"):
        self.ports = [51337]
        self.osc = []
        super().__init__(ip=ip, ports=self.ports)

    def start(self):
        """
        Start the client.
        """
        for i in range(len(self.ports)):
            try:
                self.osc.append(SimpleUDPClient(self.ip, self.ports[i]))
                print(f"Streaming OSC {i} activated on '{self.ip}:{self.ports[i]}")
            except ConnectionError:
                raise RuntimeError("Unknown error. OSC client not open.")

    @staticmethod
    def __adjust_dims(data: dict, device_to_send: dict):
        """
        Adjust the dimensions of the data to send.

        Parameters
        ----------
        data : dict
            The data to send.
        device_to_send : dict
            The device type to send the data to (emg or imu).

        Returns
        -------
        The data to send.
        """
        data_to_return = []
        for key in device_to_send:
            if key == "emg":
                emg_proc = np.array(data["emg_proc"])[:, -1:]
                emg_proc = emg_proc.reshape(emg_proc.shape[0])
                data_to_return.append(emg_proc.tolist())
            elif key == "imu":
                imu = np.array(data["imu_proc"])[:, :, -1:]
                data_to_return.append(imu.tolist())
                accel_proc = imu[:, :3, :]
                accel_proc = accel_proc.reshape(accel_proc.shape[0])
                data_to_return.append(accel_proc.tolist())
                gyro_proc = imu[:, 3:6, :]
                gyro_proc = gyro_proc.reshape(gyro_proc.shape[0])
                data_to_return.append(gyro_proc.tolist())
            else:
                raise RuntimeError(f"Unknown device ({key}) to send. Possible devices are 'emg' and 'imu'.")

        return data_to_return

    def send_data(self, data: dict, device_to_send: dict):
        """
        Send the data to the client.

        Parameters
        ----------
        data : dict
            The data to send.
        device_to_send : dict
            The device type to send the data to (emg or imu).
        """

        data = self.__adjust_dims(data, device_to_send)
        for key in device_to_send:
            if key == "emg":
                self.osc[0].send_message("/emg", data[0])
            elif key == "imu":
                idx = 1 if key in "emg" in device_to_send else 0
                self.osc[0].send_message("/imu/", data[idx])
                self.osc[0].send_message("/accel/", data[idx + 1])
                self.osc[0].send_message("/gyro/", data[idx + 2])
            else:
                raise RuntimeError(f"Unknown device ({key}) to send. Possible devices are 'emg' and 'imu'.")



"""
This file contains the functions for data processing (offline and in real-time). Both class herites
 from the GenericProcessing class.
"""

from scipy.signal import butter, lfilter, filtfilt, convolve
import numpy as np
import os
import time
from typing import Union
from ..file_io.save_and_load import save, load


class GenericProcessing:
    def __init__(self):
        """
        Initialize the GenericProcessing class.
        """
        self.bpf_lcut = 10
        self.bpf_hcut = 425
        self.lpf_lcut = 5
        self.lp_butter_order = 4
        self.bp_butter_order = 2
        self.data_rate = None
        self.process_time = []

    @staticmethod
    def _butter_bandpass(lowcut: float, highcut: float, fs: float, order: int = 5) -> tuple:
        """
        Create a butter bandpass filter.

        Parameters
        ----------
        lowcut : float
            Low cut frequency.
        highcut: float
            High cut frequency.
        fs: float
            Sampling frequency.
        order: int
            Order of the filter.

        Returns
        -------
        tuple
            Filter coefficients.
        """
        nyq = 0.5 * fs
        low = lowcut / nyq
        high = highcut / nyq
        b, a = butter(order, [low, high], btype="band")
        return b, a

    @staticmethod
    def _butter_lowpass(lowcut, fs, order=4) -> tuple:
        """
        Create a butter lowpass filter.

        Parameters
        ----------
        lowcut : float
            Low cut frequency.
        fs: float
            Sampling frequency.
        order: int
            Order of the filter.

        Returns
        -------
        tuple
            Filter coefficients.
        """
        nyq = 0.5 * fs
        low = lowcut / nyq
        b, a = butter(order, [low], btype="low")
        return b, a

    def _butter_bandpass_filter(self, data: np.ndarray, lowcut: float, highcut: float, fs: float, order: int = 5):
        """
        Apply a butter bandpass filter.

        Parameters
        ----------
        data: numpy.ndarray
            Data to filter.
        lowcut: float
            Low cut frequency.
        highcut: float
            High cut frequency.
        fs:     float
            Sampling frequency.
        order: int
            Order of the filter.

        Returns
        -------
        numpy.ndarray
            Filtered data.

        """
        b, a = self._butter_bandpass(lowcut, highcut, fs, order=order)
        y = lfilter(b, a, data)
        return y

    def butter_lowpass_filter(self, data: np.ndarray, lowcut: float, fs: float, order: int = 4) -> np.ndarray:
        """
        Apply a butter lowpass filter.

        Parameters
        ----------
        data: numpy.ndarray
            Data to filter.
        lowcut: float
            Low cut frequency.
        fs:    float
            Sampling frequency.
        order: int
            Order of the filter.

        Returns
        -------
        numpy.ndarray
            Filtered data.
        """
        b, a = self._butter_lowpass(lowcut, fs, order=order)
        y = filtfilt(b, a, data)
        return y

    @staticmethod
    def _moving_average(data: np.ndarray, window: np.ndarray, empty_ma: np.ndarray) -> np.ndarray:
        """
        Apply moving average.

        Parameters
        ----------
        data: numpy.ndarray
            Data to process.
        window: numpy.ndarray
            window to apply.
        empty_ma: numpy.ndarray
            Empty array to store the moving average.

        Returns
        -------
        numpy.ndarray
            Moving average.
        """
        for i in range(data.shape[0]):
            empty_ma[i, :] = convolve(data[i, :], window, mode="same", method="fft")
        return empty_ma

    @staticmethod
    def center(emg_data: np.ndarray, center_value: float = None) -> np.ndarray:
        """
        Center the EMG data.

        Parameters
        ----------
        emg_data : numpy.ndarray
            EMG data.
        center_value: int
            Value to center the data.

        Returns
        -------
        numpy.ndarray
            Centered EMG data.
        """
        center_value = center_value if center_value else emg_data.mean(axis=1)
        emg_centered = np.copy(emg_data)
        for i in range(emg_data.shape[0]):
            emg_centered[i, :] = emg_data[i, :] - center_value[i]
        return emg_centered

    @staticmethod
    def normalize_emg(emg_data: np.ndarray, mvc_list: list) -> np.ndarray:
        """
        Normalize EMG data.

        Parameters
        ----------
        emg_data : numpy.ndarray
            EMG data.
        mvc_list : list
            List of MVC values.

        Returns
        -------
        numpy.ndarray
            Normalized EMG data.
        """

        if len(mvc_list) == 0:
            raise RuntimeError("Please give a list of mvc to normalize the emg signal.")
        norm_emg = np.zeros((emg_data.shape[0], emg_data.shape[1]))
        for emg in range(emg_data.shape[0]):
            norm_emg[emg, :] = emg_data[emg, :] / mvc_list[emg]
        return norm_emg

    def calibration_matrix(self, data: np.ndarray, matrix: np.ndarray) -> np.ndarray:
        """
        Apply a calibration matrix to the data.

        Parameters
        ----------
        data: numpy.ndarray
            Data to calibrate.
        matrix: numpy.ndarray
            Calibration matrix.

        Returns
        -------
        numpy.ndarray
            Calibrated data.
        """
        tic = time.time()
        data_cal = np.dot(matrix, data)
        self.process_time.append(time.time() - tic)
        return data_cal

    def process_generic_signal(
        self,
        data: np.ndarray,
        norm_values: Union[list, tuple] = None,
        band_pass_filter=True,
        low_pass_filter=False,
        moving_average=True,
        centering=True,
        absolute_value=True,
        normalization=False,
        moving_average_window=200,
        **kwargs,
    ) -> np.ndarray:
        """
        Process EMG data.

        Parameters
        ----------
        data : numpy.ndarray
            EMG data.
        norm_values : list
            Values to normalize the signal.
        band_pass_filter : bool
            Apply band pass filter.
        low_pass_filter : bool
            Apply low pass filter.
        moving_average : bool
            Apply moving average.
        centering : bool
            Apply centering.
        absolute_value : bool
            Apply absolute value.
        normalization : bool
            Apply normalization.
        moving_average_window : int
            Moving average window.

        Returns
        -------
        numpy.ndarray
            Processed EMG data.
        """
        self.update_signal_processing_parameters(**kwargs)
        data_proc = data
        if band_pass_filter:
            data_proc = self._butter_bandpass_filter(data, self.bpf_lcut, self.bpf_hcut, self.data_rate)
        if centering:
            data_proc = self.center(data_proc)
        if absolute_value:
            data_proc = np.abs(data_proc)
        if low_pass_filter and moving_average:
            raise RuntimeError("Please choose between low-pass filter and moving average.")
        if low_pass_filter:
            data_proc = self.butter_lowpass_filter(data_proc, self.lpf_lcut, self.data_rate, order=self.lp_butter_order)
        elif moving_average:
            w = np.repeat(1, moving_average_window) / moving_average_window
            empty_ma = np.ndarray((data.shape[0], data.shape[1]))
            data_proc = self._moving_average(data_proc, w, empty_ma)

        if normalization:
            data_proc = self.normalize_emg(data_proc, norm_values)

        return data_proc

    def update_signal_processing_parameters(self, **kwargs):
        """
        update the signal processing parameters.

        Parameters
        ----------
        kwargs: dict
            Dictionary of parameters.
        """
        for key, value in kwargs.items():
            if key in self.__dict__:
                self.__dict__[key] = value


class RealTimeProcessing(GenericProcessing):
    def __init__(self, data_rate: Union[int, float], processing_window: int = None):
        """
        Initialize the class for real time processing.

        Parameters
        ----------
        data_rate : int
            Data rate.
        processing_window : int
            The window on which data will be processed.
        """
        super().__init__()
        self.data_rate = data_rate
        self.processing_window = processing_window if processing_window else data_rate
        self.raw_data_buffer = []
        self.processed_data_buffer = []
        self._is_one = None

    def process_emg(
        self,
        emg_data: np.ndarray,
        mvc_list: Union[list, tuple] = None,
        band_pass_filter=True,
        low_pass_filter=False,
        moving_average=True,
        centering=True,
        absolute_value=True,
        normalization=False,
        moving_average_window=200,
        **kwargs,
    ) -> np.ndarray:
        """
        Process EMG data in real-time.

        Parameters
        ----------
        emg_data : numpy.ndarray
            Temporary EMG data (nb_emg, emg_sample).
        mvc_list : list
            MVC values.
        band_pass_filter : bool
            True if apply band pass filter.
        low_pass_filter : bool
            True if apply low pass filter.
        moving_average : bool
            True if apply moving average.
        centering : bool
            True if apply centering.
        absolute_value : bool
            True if apply absolute value.
        normalization : bool
            True if apply normalization.
        moving_average_window : int
            Moving average window.

        Returns
        -------
        np.ndarray
           processed EMG data.

        """
        self.update_signal_processing_parameters(**kwargs)
        tic = time.time()
        if low_pass_filter and moving_average:
            raise RuntimeError("Please choose between low-pass filter and moving average.")
        ma_win = moving_average_window
        if ma_win > self.processing_window:
            raise RuntimeError(f"Moving average windows ({ma_win}) higher than emg windows ({self.processing_window}).")
        emg_sample = emg_data.shape[1]
        if emg_sample == 0:
            raise RuntimeError("EMG data are empty.")

        if normalization:
            if not mvc_list:
                raise RuntimeError("Please give a list of mvc to normalize the emg signal.")
            if isinstance(mvc_list, np.ndarray) is True:
                if len(mvc_list.shape) == 1:
                    quot = mvc_list.reshape(-1, 1)
                else:
                    quot = mvc_list
            else:
                quot = np.array(mvc_list).reshape(-1, 1)
        else:
            quot = [1]

        if len(self.raw_data_buffer) == 0:
            self.raw_data_buffer = emg_data
            processed_shape = self.raw_data_buffer.shape[1] if not moving_average else 1
            self.processed_data_buffer = np.zeros((emg_data.shape[0], processed_shape))

        elif self.raw_data_buffer.shape[1] < self.processing_window:
            self.raw_data_buffer = np.append(self.raw_data_buffer, emg_data, axis=1)
            if not moving_average:
                self.processed_data_buffer = np.append(
                    self.processed_data_buffer, np.zeros((emg_data.shape[0], emg_sample)), axis=1
                )
            else:
                self.processed_data_buffer = np.append(
                    self.processed_data_buffer, np.zeros((emg_data.shape[0], 1)), axis=1
                )

        else:
            self.raw_data_buffer = np.append(
                self.raw_data_buffer[:, -self.processing_window + emg_sample :], emg_data, axis=1
            )
            emg_proc_tmp = self.process_generic_signal(
                self.raw_data_buffer,
                band_pass_filter=band_pass_filter,
                centering=centering,
                absolute_value=absolute_value,
                low_pass_filter=False,
                moving_average=False,
                normalization=False,
            )
            self.processed_data_buffer = emg_proc_tmp / quot
            if low_pass_filter:
                self.processed_data_buffer = (
                    self.butter_lowpass_filter(emg_proc_tmp, self.lpf_lcut, self.data_rate, order=self.lp_butter_order)
                    / quot
                )
            elif moving_average:
                average = np.median(emg_proc_tmp[:, -ma_win:], axis=1).reshape(-1, 1)
                self.processed_data_buffer = np.append(self.processed_data_buffer[:, 1:], average / quot, axis=1)
        self.process_time.append(time.time() - tic)
        return self.processed_data_buffer

    def process_imu(
        self,
        im_data: np.ndarray,
        accel: bool = False,
        squared: bool = False,
        norm_min_bound: int = None,
        norm_max_bound: int = None,
        **kwargs,
    ) -> np.ndarray:
        """
        Process IMU data in real-time.

        Parameters
        ----------
        im_data : numpy.ndarray
            Temporary IMU data (nb_imu, im_sample).
        accel : bool
            True if current data is acceleration data to adapt the processing.
        squared : bool
            True if apply squared.
        norm_min_bound : int
            Normalization minimum bound.
        norm_max_bound : int
            Normalization maximum bound.

        Returns
        -------
        np.ndarray
            processed IMU data.
        """
        self.update_signal_processing_parameters(**kwargs)
        tic = time.time()
        if len(self.raw_data_buffer) == 0:
            if squared is not True:
                self.processed_data_buffer = np.zeros((im_data.shape[0], im_data.shape[1], 1))
            else:
                self.processed_data_buffer = np.zeros((im_data.shape[0], 1))
            self.raw_data_buffer = im_data

        elif self.raw_data_buffer.shape[2] < self.processing_window:
            self.raw_data_buffer = np.append(self.raw_data_buffer, im_data, axis=2)
            if squared is not True:
                self.processed_data_buffer = np.zeros(
                    (im_data.shape[0], im_data.shape[1], self.raw_data_buffer.shape[2])
                )
            else:
                self.processed_data_buffer = np.zeros((im_data.shape[0], self.raw_data_buffer.shape[2]))

        else:
            self.raw_data_buffer = np.append(
                self.raw_data_buffer[:, :, -self.processing_window + im_data.shape[2] :], im_data, axis=2
            )
            im_proc_tmp = self.raw_data_buffer
            average = np.mean(im_proc_tmp[:, :, -self.processing_window :], axis=2).reshape(-1, 3, 1)
            if squared:
                if accel:
                    average = abs(np.linalg.norm(average, axis=1) - 9.81)
                else:
                    average = np.linalg.norm(average, axis=1)

            if len(average.shape) == 3:
                if norm_min_bound or norm_max_bound:
                    for i in range(self.raw_data_buffer.shape[0]):
                        for j in range(self.raw_data_buffer.shape[1]):
                            if average[i, j, :] < 0:
                                average[i, j, :] = average[i, j, :] / abs(norm_min_bound)
                            elif average[i, j, :] >= 0:
                                average[i, j, :] = average[i, j, :] / norm_max_bound
                self.processed_data_buffer = np.append(self.processed_data_buffer[:, :, 1:], average, axis=2)

            else:
                if norm_min_bound or norm_max_bound:
                    for i in range(self.raw_data_buffer.shape[0]):
                        for j in range(self.raw_data_buffer.shape[1]):
                            if average[i, :] < 0:
                                average[i, :] = average[i, :] / abs(norm_min_bound)
                            elif average[i, :] >= 0:
                                average[i, :] = average[i, :] / norm_max_bound
                self.processed_data_buffer = np.append(self.processed_data_buffer[:, 1:], average, axis=1)
        self.process_time.append(time.time() - tic)
        return self.processed_data_buffer

    def get_peaks(
        self,
        new_sample: np.ndarray,
        threshold: float,
        min_peaks_interval=None,
    ) -> tuple:
        """
        Allow to get the number of peaks for an analog signal (to get cadence from treadmill for instance).

        Parameters
        ----------
        new_sample : numpy.ndarray
            New sample to add to the signal.
        threshold : float
            Threshold to detect peaks.
        min_peaks_interval : float
            Minimum interval between two peaks.

        Returns
        -------
        tuple
            Number of peaks and the processed signal.
        """
        tic = time.time()
        nb_peaks = []
        if len(new_sample.shape) == 1:
            new_sample = np.expand_dims(new_sample, 0)
        sample_proc = np.copy(new_sample)
        if not self._is_one:
            self._is_one = [False] * new_sample.shape[0]

        for i in range(new_sample.shape[0]):
            for j in range(new_sample.shape[1]):
                if new_sample[i, j] < threshold:
                    sample_proc[i, j] = 0
                    self._is_one[i] = False
                elif new_sample[i, j] >= threshold:
                    if not self._is_one[i]:
                        sample_proc[i, j] = 1
                        self._is_one[i] = True
                    else:
                        sample_proc[i, j] = 0

        if len(self.raw_data_buffer) == 0:
            self.raw_data_buffer = new_sample
            self.processed_data_buffer = sample_proc
            nb_peaks = None

        elif self.raw_data_buffer.shape[1] < self.processing_window:
            self.raw_data_buffer = np.append(self.raw_data_buffer, new_sample, axis=1)
            self.processed_data_buffer = np.append(self.processed_data_buffer, sample_proc, axis=1)
            nb_peaks = None

        else:
            self.raw_data_buffer = np.append(self.raw_data_buffer[:, new_sample.shape[1] :], new_sample, axis=1)
            self.processed_data_buffer = np.append(
                self.processed_data_buffer[:, new_sample.shape[1] :], sample_proc, axis=1
            )

        if min_peaks_interval:
            self.processed_data_buffer = RealTimeProcessing._check_and_adjust_interval(
                self.processed_data_buffer, min_peaks_interval
            )

        if isinstance(nb_peaks, list):
            nb_peaks.append(np.count_nonzero(self.processed_data_buffer))
        self.process_time.append(time.time() - tic)
        return nb_peaks, self.processed_data_buffer

    @staticmethod
    def _check_and_adjust_interval(signal, interval):
        for j in range(signal.shape[0]):
            if np.count_nonzero(signal[j, -interval:] == 1) not in [0, 1]:
                idx = np.where(signal[j, -interval:] == 1)[0]
                for i in idx[1:]:
                    signal[j, -interval:][i] = 0
        return signal

    def custom_processing(self, funct: callable, data_tmp: np.ndarray, **kwargs) -> np.ndarray:
        """
        Allow to apply a custom processing function to the data.

        Parameters
        ----------
        funct: callable
            Function to apply to the data.
        data_tmp: numpy.ndarray
            Data to process.

        Returns
        -------
        numpy.ndarray
            Processed data.
        """
        tic = time.time()
        data_tmp = funct(data_tmp, **kwargs)
        self.process_time.append(time.time() - tic)
        return data_tmp

    def get_mean_process_time(self):
        return np.mean(self.process_time)


class OfflineProcessing(GenericProcessing):
    def __init__(self, data_rate: float = None, processing_window: int = None):
        """
        Offline processing.

        Parameters
        ----------
        data_rate : float
            Data rate of the signal.
        processing_window : int
            Number of samples to process at once.
        """
        super(OfflineProcessing, self).__init__()
        self.data_rate = data_rate
        self.processing_window = processing_window

    def process_emg(self, data: np.ndarray, mvc_list: list = None, **kwargs) -> np.ndarray:
        """
        Process EMG data.

        Parameters
        ----------
        data : np.ndarray
            Raw EMG data.
        mvc_list: list
            List of MVC for each muscle.

        Returns
        -------
        np.ndarray
            Processed EMG data.

        """
        return self.process_generic_signal(data, mvc_list, **kwargs)

    @staticmethod
    def compute_mvc(
        nb_muscles: int,
        mvc_trials: np.ndarray,
        window_size: int,
        tmp_file: str = None,
        output_file: str = None,
        save_file: bool = False,
    ) -> list:
        """
        Compute MVC from several mvc_trials.

        Parameters
        ----------
        nb_muscles : int
            Number of muscles.
        mvc_trials : numpy.ndarray
            EMG data for all trials.
        window_size : int
            Size of the window to compute MVC. Usually it is 1 second so the data rate.
        tmp_file : str
            Name of the temporary file.
        output_file : str
            Name of the output file.
        save_file : bool
            If true, save the results.

        Returns
        -------
        list
            MVC for each muscle.

        """
        mvc_list_max = []
        for i in range(nb_muscles):
            mvc_temp = -np.sort(-mvc_trials, axis=1)
            if i == 0:
                mvc_list_max = mvc_temp[:, :window_size]
            else:
                mvc_list_max = np.concatenate((mvc_list_max, mvc_temp[:, :window_size]), axis=1)
        mvc_list_max = -np.sort(-mvc_list_max, axis=1)[:, :window_size]
        mvc_list_max = np.median(mvc_list_max, axis=1)

        if tmp_file:
            mat_content = load(tmp_file)
            mat_content["MVC_list_max"] = mvc_list_max
        else:
            mat_content = {"MVC_list_max": mvc_list_max, "MVC_trials": mvc_trials}

        if save_file:
            save(mat_content, output_file)
        if tmp_file:
            os.remove(tmp_file)
        return mvc_list_max



"""
This example shows how to retrieve marker data from a Vicon Nexus interface. Please note that the Vicon interface is the only implemented method capable of retrieving marker data.
First, you need to create a ViconClient object. This object will be used to connect to the Vicon system and retrieve data. Next, you need to add a set of markers to the interface. For now, only one marker set can be added.
The marker set takes the following arguments:
    - nb_markers : int
        Number of markers.
    - name : str
        Name of the marker set.
    - marker_names : Union [list, str]
        List of marker names.
    subject_name : str
        Name of the subject. If None, the subject will be the first in Nexus.
    rate: int
        Rate of the camera used to record the marker trajectories.
    unit : str
        Unit of the marker trajectories ("mm" or "m").
If you want to display the markers in a 3D scatter plot, you can add a Scatter3D plot to the interface. You can pass the size and color of the marker via size and color argument, respectively. Please see the Scatter3D documentation for more information.
Next, the data flow runs in a loop where the get_marker_set_data() function is used to retrieve the data from the interface. The data is then passed to the graph via the update() method through an array of (n_frame, n_markers, 3) where the plot parameters can be updated.
"""
from time import sleep, time
from custom_interface import MyInterface
from biosiglive import LivePlot, PlotType, ViconClient

if __name__ == "__main__":
    try_offline = True

    output_file_path = "trial_x.bio"
    if try_offline:
        interface = MyInterface(system_rate=100, data_path="abd.bio")
    else:
        interface = ViconClient(ip="localhost", system_rate=100)

    n_markers = 15
    interface.add_marker_set(
        nb_markers=n_markers, data_buffer_size=100, marker_data_file_key="markers", name="markers", rate=100, unit="mm"
    )
    marker_plot = LivePlot(name="markers", plot_type=PlotType.Scatter3D)
    marker_plot.init()
    time_to_sleep = 1 / 100
    offline_count = 0
    mark_to_plot = []
    while True:
        tic = time()
        mark_tmp, _ = interface.get_marker_set_data()
        marker_plot.update(mark_tmp[:, :, -1].T, size=0.03)
        loop_time = time() - tic
        real_time_to_sleep = time_to_sleep - loop_time
        if real_time_to_sleep > 0:
            sleep(real_time_to_sleep)



"""
This example show how to use the StreamData class to stream data from an interface, process it using several process to
finally send the data through a server to a client. Each task in a separate process to allow the streaming and the
 processing to be done in real-time. Please note that for now only a number equal to the number of cores of the computer
  is supported.
First an interface is created and device and marker_set are added to it (please refer to EMG_streming.py and
marker_streaming.py for more details).
Then a StreamData object is created. The StreamData object takes as argument the targeted frequency at which the data will
be streamed. Then the interface is added to the StreamData object. If the user want to start a server to
disseminate the data a server can be added to the StreamData object specifying the ip address and the port and the
data buffer for the device and the marker set. The data buffer is the number of frame that will be stored in the server,
it will be use if the client need a specific amount of data.
Then the streaming will be started with all the data streaming, processing and the server in seperate process. If no
 processing method is specified the data will be streamed as it is and no additional process will be started. A file can
  be specified to save the data. The data will be saved in a *.bio file at each loop of the data streaming by default or
  at the save frequency specified in the start method.
Please note that it is not yet possible to plot the data in real-time.
"""
from custom_interface import MyInterface
from biosiglive import (
    ViconClient,
    PytrignoClient,
    StreamData,
    DeviceType,
    InverseKinematicsMethods,
    RealTimeProcessingMethod,
    InterfaceType,
)

try:
    import biorbd
except ModuleNotFoundError:
    biorbd_package = False

try:
    from vicon_dssdk import ViconDataStream as VDS
except ModuleNotFoundError:
    vicon_package = False


if __name__ == "__main__":
    server_ip = "127.0.0.1"
    server_port = 50000
    interface_type = InterfaceType.Custom

    if interface_type == InterfaceType.Custom:
        interface = MyInterface(system_rate=100, data_path="abd.bio")
    elif interface_type == InterfaceType.ViconClient:
        interface = ViconClient(system_rate=100)
    elif interface_type == InterfaceType.PytrignoClient:
        interface = PytrignoClient(system_rate=100, ip="127.0.0.1")
    else:
        raise ValueError("The type of interface is not valid.")

    model_path = "model/Wu_Shoulder_Model_mod_wt_wrapp.bioMod"
    nb_electrode = 5
    interface.add_device(
        name="EMG",
        device_type=DeviceType.Emg,
        rate=2000,
        nb_channels=nb_electrode,
        device_data_file_key="emg",
        data_buffer_size=2000,
        processing_method=RealTimeProcessingMethod.ProcessEmg,
        processing_window=1,
        moving_average_window=1,
        low_pass_filter=False,
        band_pass_filter=True,
        normalization=False,
    )
    interface.get_device("EMG").process_method = RealTimeProcessingMethod.ProcessEmg
    interface.get_device("EMG").process_method_kwargs = {
        "processing_window": 2000,
        "low_pass_filter": False,
        "band_pass_filter": True,
    }

    interface.add_marker_set(
        name="markers",
        rate=100,
        data_buffer_size=100,
        kinematics_method=InverseKinematicsMethods.BiorbdKalman,
        model_path=model_path,
        marker_data_file_key="markers",
        nb_markers=16,
    )
    data_streaming = StreamData(stream_rate=100)
    data_streaming.add_interface(interface)
    data_streaming.add_server(server_ip, server_port, device_buffer_size=20, marker_set_buffer_size=1)
    data_streaming.start(save_streamed_data=True, save_path="data_streamed")



"""
This file contains all the plot functions to plot the data in live or offline.
"""
try:
    import pyqtgraph as pg
    import pyqtgraph.opengl as gl
    from PyQt5.QtWidgets import QProgressBar
except ModuleNotFoundError:
    pass
import numpy as np
from typing import Union
import matplotlib.pyplot as plt
from math import ceil
from ..enums import PlotType
import time


# TODO: Add plot class to be able to plot several curves in the same plot. And do several plot in the same app.
class LivePlot:
    def __init__(
        self,
        nb_subplots: int = 1,
        plot_type: Union[PlotType, str] = PlotType.Curve,
        name: str = None,
        channel_names: list = None,
        rate: int = None,
    ):
        """
        Initialize the plot class.

        Parameters
        ----------
        plot_type : Union[PlotType, str]
            type of the plot (curve, progress bar, 3D scatter, skeleton...).
        name : str
            name of the plot
        channel_names : list
            list of the channel names (subplots names)
        nb_subplots : int
            number of subplots
        """
        if isinstance(plot_type, str):
            if plot_type not in [t.value for t in PlotType]:
                raise ValueError("Plot type not recognized")
            plot_type = PlotType(plot_type)
        self.plot_type = plot_type

        if nb_subplots and channel_names:
            if len(channel_names) != nb_subplots:
                raise ValueError("The number of subplots is not equal to the number of channel names")

        self.channel_names = channel_names
        self.figure_name = name
        self.nb_subplot = nb_subplots
        self.rate = rate
        self.layout = None
        self.app = None
        self.viz = None
        self.resize = (400, 400)
        self.move = (0, 0)
        self.plot_windows = None
        self.plot_buffer = None
        self.msk_model = None
        self.last_plot = None
        self.once_update = False
        self.plots = []
        self.curves = []
        self.ptr = []
        self.unit = ""
        self.size_to_append = []

    def init(
        self,
        plot_windows: Union[int, list] = None,
        **kwargs,
    ):
        """
        This function is used to initialize the qt app.

        Parameters
        ----------
        plot_windows: Union[int, list]
            The number of frames ti plot. If is a list, the number of frames to plot for each subplot.
        """
        self.plot_buffer = [None] * self.nb_subplot
        if isinstance(plot_windows, int):
            plot_windows = [plot_windows] * self.nb_subplot
        self.plot_windows = plot_windows
        if self.plot_type == PlotType.Curve:
            self._init_curve(self.figure_name, self.channel_names, self.nb_subplot, **kwargs)
        elif self.plot_type == PlotType.ProgressBar:
            self._init_progress_bar(self.figure_name, self.nb_subplot, **kwargs)
        elif self.plot_type == PlotType.Scatter3D:
            if self.nb_subplot != 1:
                raise ValueError("The number of subplots should be 1 for 3DScatter plot.")
            self._init_3d_scatter(self.figure_name, **kwargs)

        elif self.plot_type == PlotType.Skeleton:
            self.viz = self._init_skeleton(**kwargs)

        else:
            raise ValueError(f"The plot type ({self.plot_type}) is not supported.")

    def update(self, data: Union[np.ndarray, list], **kwargs):
        """
        This function is used to update the qt app.

        Parameters
        ----------
        data: Union[np.ndarray, list]
            The data to plot. If it is a list, the data to plot for each subplot.
        """
        update = True
        if self.plot_type != PlotType.Scatter3D and self.plot_type != PlotType.Skeleton:
            if isinstance(data, list):
                if len(data) != self.nb_subplot:
                    raise ValueError("The number of subplots is not equal to the number of data.")
                for d in data:
                    if isinstance(d, np.ndarray):
                        if len(d.shape) != 2:
                            raise ValueError("The data should be a 2D array.")
                    else:
                        raise ValueError("The data should be a 2D array.")
            if isinstance(data, np.ndarray):
                data_mat = data
                data = []
                if data_mat.shape[0] != self.nb_subplot:
                    raise ValueError("The number of subplots is not equal to the number of data.")
                for d in data_mat:
                    data.append(d[np.newaxis, :])

        if self.plot_windows:
            for i in range(self.nb_subplot):
                if self.plot_buffer[i] is None:
                    self.plot_buffer[i] = data[i][..., -self.plot_windows[i] :]
                    if self.plot_buffer[i].shape[1] < self.plot_windows[i]:
                        size = self.plot_windows[i] - self.plot_buffer[i].shape[1]
                        self.plot_buffer[i] = np.append(
                            np.zeros((self.plot_buffer[i].shape[0], size)), self.plot_buffer[i], axis=-1
                        )
                elif self.plot_buffer[i].shape[1] < self.plot_windows[i]:
                    self.plot_buffer[i] = np.append(self.plot_buffer[i], data[i], axis=-1)
                elif self.plot_buffer[i].shape[1] >= self.plot_windows[i]:
                    size = data[i].shape[1]
                    self.plot_buffer[i] = np.append(self.plot_buffer[i][..., size:], data[i], axis=-1)
            data = self.plot_buffer
        if self.rate and self.once_update:
            plot_time = time.time() - self.last_plot
            if plot_time != 0 and 1 / plot_time > self.rate:
                update = False
            else:
                update = True
        if update:
            self.once_update = True
            if self.plot_type == PlotType.ProgressBar:
                self._update_progress_bar(data)
            elif self.plot_type == PlotType.Curve:
                self._update_curve(data)
            elif self.plot_type == PlotType.Skeleton:
                self._update_skeleton(data, self.viz)
            elif self.plot_type == PlotType.Scatter3D:
                self._update_3d_scatter(data, **kwargs)
            else:
                raise ValueError(f"The plot type ({self.plot_type}) is not supported.")
            self.last_plot = time.time()

    def _init_curve(
        self,
        figure_name: str = "Figure",
        subplot_labels: Union[list, str] = None,
        nb_subplot: int = None,
        x_labels: Union[list, str] = None,
        y_labels: Union[list, str] = None,
        grid: bool = True,
        colors: Union[list, tuple] = None,
    ):
        """
        This function is used to initialize the curve plot.

        Parameters
        ----------
        figure_name: str
            The name of the figure.
        subplot_labels: Union[list, str]
            The labels of the subplots.
        nb_subplot: int
            The number of subplot.
        x_labels: Union[list, str]
            The labels of the x axis.
        y_labels: Union[list, str]
            The labels of the y axis.
        grid: bool
            If True, the grid is displayed.
        colors: Union[list, tuple]
            The colors of the curves.
        """
        # --- Curve graph --- #
        self.app = pg.mkQApp("Curve_plot")
        pg.setConfigOption("background", "w")
        pg.setConfigOption("foreground", "k")
        self.win = pg.GraphicsLayoutWidget(show=True)
        self.win.setWindowTitle(figure_name)
        nb_line = 4
        nb_col = ceil(nb_subplot / nb_line)
        line_count = 0
        self.win.resize(self.resize[0], self.resize[1])
        self.win.move(self.move[0], self.move[1])
        if colors:
            if isinstance(colors, tuple):
                colors = [colors] * nb_subplot
            elif isinstance(colors, list):
                if len(colors) != nb_subplot:
                    raise ValueError("The number of colors is not equal to the number of subplots.")
        else:
            colors = [(0, 128, 232)] * nb_subplot  # Blue
        if not x_labels:
            x_labels = ["Frames"] * nb_subplot
        else:
            if isinstance(x_labels, str):
                x_labels = [x_labels] * nb_subplot
            elif isinstance(x_labels, list):
                if len(x_labels) != nb_subplot:
                    raise ValueError("The number of x labels is not equal to the number of subplots.")

        if not y_labels:
            y_labels = ["Amplitude"] * nb_subplot
        else:
            if isinstance(y_labels, str):
                y_labels = [y_labels] * nb_subplot
            elif isinstance(y_labels, list):
                if len(y_labels) != nb_subplot:
                    raise ValueError("The number of y labels is not equal to the number of subplots.")
        if not subplot_labels:
            subplot_labels = [f"Subplot {i}" for i in range(nb_subplot)]
        else:
            if isinstance(subplot_labels, list):
                if len(subplot_labels) != nb_subplot:
                    raise ValueError("The number of subplot labels is not equal to the number of subplots.")
        for subplot in range(nb_subplot):
            self.ptr.append(0)
            self.size_to_append.append(0)
            if line_count == nb_col:
                self.win.nextRow()
                line_count = 0
            self.plots.append(self.win.addPlot(title=subplot_labels[subplot]))
            self.plots[-1].setDownsampling(mode="peak")
            self.plots[-1].setClipToView(False)
            self.curves.append(self.plots[-1].plot([], pen=colors[subplot], name="Blue curve"))
            self.plots[-1].setLabel("bottom", x_labels[subplot])
            self.plots[-1].setLabel("left", y_labels[subplot])
            self.plots[-1].showGrid(x=grid, y=grid)
            line_count += 1

    def _init_progress_bar(
        self,
        figure_name: str = "Figure",
        nb_subplot: int = None,
        bar_graph_max_value: Union[int, list] = 100,
        unit: Union[str, list] = "",
    ):
        """
        This function is used to initialize the curve plot.

        Parameters
        ----------
        figure_name: str
            The name of the figure.
        nb_subplot: int
            The number of subplot.
        bar_graph_max_value: int or list
            The maximum value of the bar graph.
        unit: str or list
            The unit of the bar graph.
        """
        # --- Progress bar graph --- #
        if isinstance(unit, str):
            self.unit = [unit] * nb_subplot
        self.layout, self.app = self._init_layout(figure_name, resize=self.resize, move=self.move)
        row_count = 0
        if bar_graph_max_value is None:
            bar_graph_max_value = [100] * nb_subplot
        if isinstance(bar_graph_max_value, int):
            bar_graph_max_value = [bar_graph_max_value] * nb_subplot
        for plot in range(nb_subplot):
            self.plots.append(QProgressBar())
            self.plots[-1].setMaximum(bar_graph_max_value[plot])
            self.layout.addWidget(self.plots[-1], row=plot, col=0)
            self.layout.show()
            row_count += 1

    def _init_3d_scatter(
        self,
        figure_name: str = "Figure",
        colors: Union[list, tuple] = (1.0, 0.0, 0.0, 0.5),
        size: Union[int, list] = 0.03,
    ):
        """
        This function is used to initialize the 3d scatter plot.

        Parameters
        ----------
        figure_name: str
            The name of the figure.
        colors: Union[list, tuple]
            The color of the scatter.
        size: Union[int, list]
            The size of the scatters.
        """
        # --- 3D scatter graph --- #
        self.app = pg.mkQApp("3D_scatter_plot")
        w = gl.GLViewWidget()
        w.opts["bgcolor"] = (0.2, 0.2, 0.2, 10)
        w.opts["distance"] = 8
        w.show()
        w.setWindowTitle(figure_name)
        g = gl.GLGridItem()
        # g.setColor((1, 1, 1, 100))
        w.addItem(g)
        pos = np.zeros((1, 3))
        self.plots.append(gl.GLScatterPlotItem(pos=pos, color=colors, size=size, pxMode=False))
        w.addItem(self.plots[-1])

    def _update_3d_scatter(
        self,
        data: Union[np.ndarray, list],
        colors: Union[list, tuple] = (0, 1.0, 0.0, 50),
        size: Union[list, float] = 0.03,
    ):
        """
        This function is used to update the 3d scatter plot.

        Parameters
        ----------
        data: np.ndarray
            The data to plot. (N, 3)
        colors: Union[list, tuple]
            The color of the scatter.
        size: float
            The size of the scatter.
        """
        if isinstance(data, np.ndarray):
            if len(data.shape) != 2:
                raise ValueError("The data must be a 2D array.")
            if data.shape[1] != 3:
                raise ValueError("The data must be a (N, 3) array.")
        if isinstance(colors, list):
            if len(colors) != len(data):
                raise ValueError("The number of colors is not equal to the number of data.")
        if isinstance(size, list):
            if len(size) != len(data):
                raise ValueError("The number of size is not equal to the number of data.")
        for plot in self.plots:
            plot.setData(pos=data, color=colors, size=size)
        self.app.processEvents()

    def _update_curve(self, data: list):
        """
        This function is used to update the curve plot.

        Parameters
        ----------
        data: list
            The data to plot.
        """
        if len(data) != len(self.curves):
            raise ValueError(
                f"The number of data ({len(data)}) is different from the number of curves ({len(self.curves)})."
            )
        for i in range(len(data)):
            if self.ptr[i] == 0:
                self.size_to_append[i] = data[i].shape[1]
            self.ptr[i] += self.size_to_append[i] * 2
            self.curves[i].setData(data[i][0, :])
            # self.curves[i].setPos(self.ptr[i], 0)
        self.app.processEvents()

    def _update_progress_bar(self, data: list):
        """
        This function is used to update the progress bar plot.

        Parameters
        ----------
        data: list
            The data to plot.
        """

        if self.channel_names and len(self.channel_names) != len(data):
            raise RuntimeError(
                f"The length of Subplot labels ({len(self.channel_names)}) is different than"
                f" the first dimension of your data ({len(data)})."
            )

        for i in range(len(data)):
            value = np.mean(data[i][0, :])
            self.plots[i].setValue(int(value))
            name = self.channel_names[i] if self.channel_names else f"plot_{i}"
            self.plots[i].setFormat(f"{name}: {int(value)} {self.unit[i]}")
        self.app.processEvents()

    @staticmethod
    def _update_skeleton(data: list, viz):
        """
        This function is used to update the skeleton plot.

        Parameters
        ----------
        data  : list
            The data to plot. list of length degree of freedom.
        viz: Viz3D
            The plot.

        Returns
        -------

        """
        viz.set_q(data[:, -1], refresh_window=True)

    @staticmethod
    def _init_skeleton(**kwargs):
        try:
            import bioviz
        except ImportError:
            raise ImportError("Please install bioviz (github.com/pyomeca/bioviz) to use the skeleton plot.")
        if not "model_path" in kwargs or "model" in kwargs:
            raise ValueError(
                "You must provide a model_path or a model to use the skeleton plot through"
                " the keyword arguments 'model_path' or 'model' respectively."
            )
        plot = bioviz.Viz(**kwargs)
        return plot

    @staticmethod
    def _init_layout(figure_name: str = "Figure", resize: tuple = (400, 400), move: tuple = (0, 0)):
        """
        This function is used to initialize the qt app layout.

        Parameters
        ----------
        figure_name: str
            The name of the figure.
        resize: tuple
            The size of the figure.
        move: tuple
            The position of the figure.

        Returns
        -------
        layout: QVBoxLayout
            The layout of the qt app.
        app: QApplication
            The qt app.

        """
        app = pg.mkQApp(figure_name)
        layout = pg.LayoutWidget()
        layout.resize(resize[0], resize[1])
        layout.move(move[0], move[1])
        return layout, app

    def disconnect(self):
        self.app.disconnect()
        try:
            self.app.closeAllWindows()
        except RuntimeError:
            pass


class OfflinePlot:
    """
    This class is used to plot data offline.
    """

    @staticmethod
    def multi_plot(
        data: Union[list, np.ndarray],
        x: Union[list, np.ndarray] = None,
        nb_column: int = None,
        y_label: str = None,
        x_label: str = None,
        legend: Union[list, str] = None,
        subplot_title: Union[str, list] = None,
        figure_name: str = None,
    ):
        """
        This function is used to plot multiple data in one figure.

        Parameters
        ----------
        data: list or np.ndarray
            The data to plot.
        x: list or np.ndarray
            The x-axis data.
        nb_column: int
            The number of columns in the figure.
        y_label: str
            The y-axis label.
        x_label: str
            The x-axis label.
        legend: list or str
            The legend of the data.
        subplot_title: str or list
            The title of the subplot.
        figure_name: str
            The name of the figure.
        """

        if not isinstance(data, list):
            data = [data]
        nb_data = len(data)
        plt.figure(figure_name)
        size_police = 12
        if nb_column:
            col = nb_column
        else:
            col = data[0].shape[0] if data[0].shape[0] <= 4 else 4
        line = ceil(data[0].shape[0] / col)
        if isinstance(legend, str):
            legend = [legend]
        for i in range(data[0].shape[0]):
            plt.subplot(line, col, i + 1)
            if y_label and i % 4 == 0:
                plt.ylabel(y_label, fontsize=size_police)
            if x_label:
                plt.xlabel(x_label, fontsize=size_police)
            for j in range(nb_data):
                if legend:
                    legend_tmp = legend[j]
                else:
                    legend_tmp = None
                if x is not None:
                    plt.plot(x, data[j][i, :], label=legend_tmp)
                else:
                    plt.plot(data[j][i, :], label=legend_tmp)

            plt.legend()
            if subplot_title:
                plt.title(subplot_title[i], fontsize=size_police)
        plt.show()



from biosiglive import load, save
import numpy as np
import pytest
import os


@pytest.mark.parametrize("rt", [True, False])
def test_save_and_load(rt):
    np.random.seed(50)
    data = {
        "data_np": np.random.rand(2, 20),
        "data_list": [5, 8],
        "data_dict": {"data_np": np.random.rand(2, 20)},
        "data_int": 1,
    }
    shape = 50 if rt else 1
    shape_np = 50 * 20 if rt else 20
    i = 0
    while i != 50:
        if rt:
            save(data, "test")
        i += 1
    if not rt:
        save(data, "test")
    data_loaded = load("test.bio")
    os.remove("test.bio")
    np.testing.assert_almost_equal(len(list(data_loaded.keys())), len(list(data.keys())))
    np.testing.assert_almost_equal(data_loaded["data_np"].shape, (2, shape_np))
    np.testing.assert_almost_equal(len(data_loaded["data_int"]), shape)
    np.testing.assert_almost_equal(len(data_loaded["data_list"]), shape)
    np.testing.assert_almost_equal(len(data_loaded["data_dict"]), shape)



"""
This example shows how to use the plot in real time inside a loop.
Here three different plots are used:
- Curve plot : allows to plot a curve on a number of frame plot_windows.
- progress barplot : Plot a progress bar with a maximum value to fit the filled area in the bar. A unit can also be sent to the function to improve the plot.
-3D scatter plot: Plots the scatter in a 3D space. The size and color can be changed using the size and color arguments, respectively.
The initialization is the same for each plot. The number of subplots can be specified as well as their names. The type of plot must be specified using the PlotType enum class. Finally, the targeted plot frequency can be defined, it will slow down the update and can therefore improve performance or visibility (for example, plotting bars at 100 Hz is difficult to understand).
Once initialized each plot can be updated in a loop using some data (here random ones).
"""

import numpy as np
from time import time, sleep
from biosiglive import LivePlot, PlotType


if __name__ == "__main__":
    plot_curve = LivePlot(
        name="curve",
        rate=100,
        plot_type=PlotType.Curve,
        nb_subplots=4,
        channel_names=["1", "2", "3", "4"],
    )
    plot_curve.init(plot_windows=1000, y_labels=["Strikes", "Strikes", "Force (N)", "Force (N)"])
    plot = LivePlot(
        name="bar",
        rate=10,
        plot_type=PlotType.ProgressBar,
        nb_subplots=4,
        channel_names=["1", "2", "3", "4"],
    )
    plot.init(bar_graph_max_value=100, unit="N")
    plot_scatter = LivePlot(
        name="scatter",
        rate=50,
        plot_type=PlotType.Scatter3D,
    )
    plot_scatter.init()
    while True:
        tic = time()
        data = np.random.rand(3, 4, 1) * 100
        plot.update(data[0, :, :])
        plot_curve.update(data[0, :, :])
        plot_scatter.update(data[:, :, -1].T / 100, size=0.1)
        sleep((1 / 100))



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys

sys.path.insert(0, os.path.abspath("../../"))

project = "biosiglive"
copyright = "2023, aceglia"
author = "aceglia"
release = "2.0.1"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "myst_parser",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]

source_suffix = {
    ".rst": "restructuredtext",
    ".txt": "markdown",
    ".md": "markdown",
}



"""
This example shows how to build a custom interface from the GenericInterface class.
The custom interface is used to distribute data from a source that is not supported by the library.
If you create a new interface, you are welcome to contribute to the library by creating a pull request following the contributing.md directives.
Here, the interface aims to distribute data from a file, each time the get data function is called, the data from the file will be sent. To see the functions to implement in this class, please refer to the GenericInterface class.
For this custom interface, you need to provide a file with extension .bio (biosiglive file format, please refer to the save_and_load.py save example) as argument. When creating the device, you must specify the device key in the dictionary read from the file using the *_data_file_key argument.
Please see any example to see how to use this interface.
"""

from typing import Union
import numpy as np
from biosiglive import (
    GenericInterface,
    DeviceType,
    InterfaceType,
    RealTimeProcessingMethod,
    OfflineProcessingMethod,
    InverseKinematicsMethods,
    load,
)


class MyInterface(GenericInterface):
    def __init__(self, system_rate: float = 100, data_path: str = None):
        super().__init__(system_rate=system_rate, interface_type=InterfaceType.Custom)
        self.offline_data = None
        if data_path:
            self.offline_data = load(data_path)
        self.device_data_key = []
        self.marker_data_key = []
        self.c = 0
        self.d = 0

    def add_device(
        self,
        nb_channels: int = 1,
        device_type: Union[DeviceType, str] = DeviceType.Emg,
        name: str = None,
        data_buffer_size: int = None,
        rate: float = 2000,
        device_range: tuple = None,
        device_data_file_key: str = None,
        processing_method: Union[RealTimeProcessingMethod, OfflineProcessingMethod] = None,
        **process_kwargs,
    ):
        device_tmp = self._add_device(
            nb_channels, device_type, name, rate, device_range, processing_method, **process_kwargs
        )
        if data_buffer_size:
            device_tmp.data_window = data_buffer_size
        self.devices.append(device_tmp)
        if self.offline_data is not None:
            if not device_data_file_key:
                raise ValueError("You need to specify the device data file key.")
        self.device_data_key.append(device_data_file_key)

    def add_marker_set(
        self,
        nb_markers: int = 3,
        name: str = None,
        data_buffer_size: int = None,
        marker_names: Union[str, list] = None,
        rate: float = 100,
        unlabeled: bool = False,
        subject_name: str = None,
        unit: str = "m",
        marker_data_file_key: str = None,
        kinematics_method: InverseKinematicsMethods = None,
        **kin_method_kwargs,
    ):
        if len(self.marker_sets) != 0:
            raise ValueError("Only one marker set can be added for now.")

        markers_tmp = self._add_marker_set(
            nb_markers=nb_markers,
            name=name,
            marker_names=marker_names,
            rate=rate,
            unlabeled=unlabeled,
            unit=unit,
            kinematics_method=kinematics_method,
            **kin_method_kwargs,
        )
        if self.offline_data is not None:
            if not marker_data_file_key:
                raise ValueError("You need to specify the marker data file key.")
        markers_tmp.subject_name = subject_name
        markers_tmp.data_windows = data_buffer_size
        self.marker_sets.append(markers_tmp)
        self.marker_data_key.append(marker_data_file_key)

    def get_device_data(self, device_name: Union[str, list] = "all", channel_names: str = None, **kwargs):
        """
        Get random data.
        Parameters
        ----------
        device_name: str or list
            Name of the device or list of devices names.
        channel_names: str
            Name of the channel.
        Returns
        -------
        device_data: list
            All asked device data.
        """
        devices = []
        all_device_data = []
        if not isinstance(device_name, list):
            device_name = [device_name]

        if device_name != ["all"]:
            for d, device in enumerate(self.devices):
                if device.name == device_name[d]:
                    devices.append(device)
        else:
            devices = self.devices

        for d, device in enumerate(devices):
            if self.offline_data:
                device.new_data = self.offline_data[self.device_data_key[d]][
                    : device.nb_channels, self.c : self.c + device.sample
                ]
                if abs(self.c + device.sample - self.offline_data[self.device_data_key[d]].shape[1]) > device.sample:
                    self.c = self.c + device.sample
                else:
                    self.c = 0
            else:
                device.new_data = np.random.rand(device.nb_channels, device.sample)
            device.append_data(device.new_data)
            all_device_data.append(device.new_data)

        if len(all_device_data) == 1:
            return all_device_data[0]
        return all_device_data

    def get_marker_set_data(self, marker_set_name: Union[list, str] = None, **kwargs):
        """
        Get the markers data from Vicon.
        Parameters
        ----------
        marker_set_name: Union[list, str]
            List of the marker set.
        Returns
        -------
        markers_data: list
            All asked markers data.
        """
        occluded = []
        all_markers_data = []
        all_occluded_data = []
        if not marker_set_name:
            nb_markers = self.marker_sets[0].nb_channels
            marker_set_name = self.marker_sets[0].name
        else:
            nb_markers = len(marker_set_name)
        if not isinstance(marker_set_name, list):
            marker_set_name = [marker_set_name]
        for m, marker in enumerate(self.marker_sets):
            coef = 1 if marker.unit == "m" else 0.001
            if marker.name == marker_set_name[m]:
                if self.offline_data:
                    marker.new_data = (
                        self.offline_data[self.marker_data_key[m]][
                            :, : marker.nb_channels, self.d : self.d + marker.sample
                        ]
                        * coef
                    )
                    if (
                        abs(self.d + marker.sample - self.offline_data[self.marker_data_key[m]].shape[2])
                        > marker.sample
                    ):
                        self.d = self.d + marker.sample
                    else:
                        self.d = 0
                else:
                    marker.new_data = np.random.rand(3, nb_markers, marker.sample)
                marker.append_data(marker.new_data)
                all_markers_data.append(marker.new_data)
                all_occluded_data.append(occluded)
        if len(all_markers_data) == 1:
            return all_markers_data[0], all_occluded_data[0]
        return all_markers_data, all_occluded_data

    def get_kinematics_from_markers(
        self,
        marker_set_name: str,
        nb_dof: int = None,
        model_path: str = None,
        method: Union[InverseKinematicsMethods, str] = InverseKinematicsMethods.BiorbdLeastSquare,
        custom_func: callable = None,
        get_markers_data: bool = False,
        **kwargs,
    ):
        if get_markers_data:
            self.get_marker_set_data(marker_set_name)
        if not self.offline_data:
            if not nb_dof:
                raise Exception("You need to specify the number of dof")
            return np.random.rand(nb_dof, 1), np.random.rand(nb_dof, 1)
        else:
            marker_set_idx = [i for i, m in enumerate(self.marker_sets) if m.name == marker_set_name][0]
            return self.marker_sets[marker_set_idx].get_kinematics(
                model_path, method, custom_func=custom_func, **kwargs
            )


if __name__ == "__main__":
    interface = MyInterface(system_rate=100, data_path="abd.bio")
    interface.add_device(
        nb_channels=8, device_type=DeviceType.Emg, name="My EMG device", rate=2000, device_data_file_key="emg"
    )
    interface.add_marker_set(
        nb_markers=16,
        name="My markers",
        unit="mm",
        rate=100,
        data_buffer_size=100,
        marker_data_file_key="markers",
        model_path="model/Wu_Shoulder_Model_mod_wt_wrapp.bioMod",
        kinematics_method=InverseKinematicsMethods.BiorbdKalman,
    )

    print(interface.get_kinematics_from_markers("My markers", get_markers_data=True))



"""
Make sure before using this example that a server is running, if you don't have one, please run the "server.py" example.
This example shows how to get data from a biosiglive server using the biosiglive client.
Here the client connection is used as an interface of the GenericInterface class.
To initialize the interface, you must give the server IP address (IP and port) and the target frequency at which the data will be broadcast. Please use "127.0.0.1" instead of "localhost" if you are using the server on the same computer as the client, to avoid connection problems.
Once the interface is initialized, you can add devices to it.
The device arguments are:
- nb_channels: int
         Number of device channels.
- command_name: str
         Name of the command to send to the server to get the data. The command must be the same as the one used in the data dictionary given to the server.
- name: str
         Device name.
- rate: float
         Device frequency.

A set of markers can also be added to the interface. The arguments of the marker set are as follows
- number of markers: int
         Number of markers in the marker set.
     name: str
         Marker set name.
- rate: float
         Marker set frequency.
- command_name: str
         Name of the command to send to the server to get the data. The command must be the same as the one used in the data dictionary given to the server.

Once the devices and marker sets have been added to the interface, you can start the interface.
After initializing the interface, the data exchange is done in a loop.
What is important here is that data can be retrieved from the server in two ways:
- using the get_data_from_server() method of the interface. In this method, you specify the command to send to the server to get the data. The command must be the same as the one used in the data dictionary given to the server. You can specify the number of frames to get from the server (last, last 10, etc., or all frames). You can also specify a downsampling factor to reduce the number of frames. The last two arguments aim to reduce the amount of data to get from the server. This method is used to have all server data at the same time to synchronize data.
- by using the get-device_data or get_marker_set_data methods of the interface. In this case, you do not need to specify the command because it has already been defined when adding the device or marker to the interface. In this case, you can either get the data directly from the server using the get from server argument, or get the data from the last call of the get_data_from_server() method. If you want synchronized data, you must call the get_data_from_server() method before retrieving data from the device or marker set. Please note that if you have read data from the server once, you cannot read it again. If you want to read data through multiple clients, you must use the same number of servers.
"""

from biosiglive import (
    TcpClient,
    DeviceType,
)
import time

if __name__ == "__main__":
    server_ip = "127.0.0.1"
    server_port = 50000
    tcp_client = TcpClient(server_ip, server_port, read_frequency=100)
    tcp_client.add_device(
        5, command_name="proc_device_data", device_type=DeviceType.Emg, name="processed EMG", rate=2000
    )
    tcp_client.add_marker_set(15, name="markers", rate=100, command_name="marker_set_data")
    while True:
        tcp_client.get_data_from_server(
            command=["proc_device_data", "marker_set_data"], nb_frame_to_get=10, down_sampling={"proc_device_data": 5}
        )
        data = tcp_client.get_device_data(device_name="processed EMG", get_from_server=False)
        data_mark = tcp_client.get_marker_set_data(marker_set_name="markers", get_from_server=False)
        time.sleep(0.01)
        print(data)
        print(data_mark)



"""
Groups the different enums used in the program. it is a good place to start to check what's available.
"""
from enum import Enum


class InterfaceType(Enum):
    """
    The different types of interfaces that can be used.
    """

    ViconClient = "vicon_client"
    PytrignoClient = "pytrigno_client"
    TcpClient = "tcp_client"
    Custom = "custom"


class DeviceType(Enum):
    """
    The different types of devices that can be used.
    """

    Emg = "emg"
    Imu = "imu"
    Generic = "generic"
    ForcePlate = "force_plate"


class MarkerType(Enum):
    """
    The different types of markers that can be used.
    """

    Labeled = "labeled"
    Unlabeled = "unlabeled"


class PlotType(Enum):
    """
    The different types of plots that can be used.
    """

    Curve = "curve"
    ProgressBar = "progress_bar"
    Skeleton = "skeleton"
    Scatter3D = "scatter3d"


class InverseKinematicsMethods(Enum):
    """
    The different types of inverse kinematics methods that can be used.
    """

    BiorbdKalman = "biorbd_kalman"
    BiorbdLeastSquare = "biorbd_least_square"
    Custom = "custom"


class RealTimeProcessingMethod(Enum):
    """
    The different types of real time processing methods that can be used.
    """

    ProcessEmg = "process_emg"
    ProcessImu = "process_imu"
    CalibrationMatrix = "calibration_matrix"
    GetPeaks = "get_peaks"
    Custom = "custom"
    ProcessGenericSignal = "process_generic_signal"


class OfflineProcessingMethod(Enum):
    """
    The different types of offline processing methods that can be used.
    """

    ProcessEmg = "process_emg"
    CalibrationMatrix = "calibration_matrix"
    ComputeMvc = "compute_mvc"



import pytest
from biosiglive import PytrignoClient, DeviceType
import numpy as np


@pytest.mark.parametrize("init", [True, False])
@pytest.mark.parametrize("get_frame", [True, False])
def test_interface(init, get_frame):
    # Create interface
    interface = PytrignoClient(system_rate=100, init_now=init)

    # Add device
    interface.add_device(name="EMG", device_type=DeviceType.Emg, nb_channels=5, device_range=(3, 8))

    if not init:
        interface.init_client()
    i = 0
    device_data = None
    while i != 50:
        # Get data
        if not get_frame:
            interface.get_frame()
        device_data = interface.get_device_data(get_frame=get_frame)
        i += 1

    np.testing.assert_almost_equal(device_data.shape[0], 5)



import numpy as np
from .generic_interface import GenericInterface
from ..enums import DeviceType, InterfaceType, RealTimeProcessingMethod, OfflineProcessingMethod
from typing import Union

try:
    import pytrigno
except ModuleNotFoundError:
    pass


class PytrignoClient(GenericInterface):
    """
    Class to wrap the Trigno community SDK.
    """

    def __init__(self, system_rate=100, ip: str = "127.0.0.1", init_now: bool = True):
        """
        Initialize the interface.

        Parameters
        ----------
        system_rate: int
            Rate of the system.
        ip: str
            IP address of the Trigno system.
        init_now: bool
            Initialize the client the same time that it is created.
        """
        super(PytrignoClient, self).__init__(
            ip=ip, interface_type=InterfaceType.PytrignoClient, system_rate=system_rate
        )
        self.address = ip
        self.devices = []
        self.imu = []
        self.markers = []

        self.emg_client, self.imu_client = [], []
        self.is_frame = False
        self.is_initialized = False
        self.init_now = init_now

    def add_device(
        self,
        nb_channels: int,
        device_type: Union[DeviceType, str] = DeviceType.Emg,
        data_buffer_size: int = None,
        name: str = None,
        rate: float = 2000,
        device_range: tuple = None,
        processing_method: Union[RealTimeProcessingMethod, OfflineProcessingMethod] = None,
        **process_kwargs,
    ):
        """
        Add a device to the Pytrigno system.

        Parameters
        ----------
        nb_channels: int
            Number of channels of the device.
        device_type: Union[DeviceType, str]
            Type of the device.
        data_buffer_size: int
            Size of the buffer for the device.
        name: str
            Name of the device.
        rate: float
            Rate of the device.
        device_range: tuple
            Range of the device. Number of selected channels. If None, all channels are selected (0, 16).
        processing_method : Union[RealTimeProcessingMethod, OfflineProcessingMethod]
            Method used to process the data.
        **process_kwargs
            Keyword arguments for the processing method.
        """
        if not device_range:
            device_range = (0, nb_channels - 1)
        device_tmp = self._add_device(
            nb_channels, device_type, name, rate, device_range, processing_method, **process_kwargs
        )
        device_tmp.interface = self.interface_type
        device_tmp.data_windows = data_buffer_size
        self.devices.append(device_tmp)
        if isinstance(device_type, str):
            if device_type not in [t.value for t in DeviceType]:
                raise ValueError("The type of the device is not valid.")
            device_type = DeviceType(device_type)
        if device_type != DeviceType.Emg and device_type != DeviceType.Imu:
            raise RuntimeError("Device type must be 'emg' or 'imu' with pytrigno.")
        if self.init_now:
            self.init_client()

    def get_device_data(
        self, device_name: str = "all", channel_idx: Union[int, list] = (), get_frame: bool = True
    ) -> np.ndarray:
        """
        Get data from the device.

        Parameters
        ----------
        device_name : str
            Name of the device.
        channel_idx : Union[int, list]
            Index of the channel.
        get_frame : bool
            Get data from device. If False, use the last data acquired.

        Returns
        -------
        data : list
            Data from the device.
        """
        if not self.is_initialized:
            raise RuntimeError("Client is not initialized. Please call init_client() first.")
        devices = []
        device_data = []
        all_device_data = []
        if channel_idx and not isinstance(channel_idx, list):
            channel_idx = [channel_idx]

        if device_name and not isinstance(device_name, list):
            device_name = [device_name]

        if device_name != "all":
            for d, device in enumerate(self.devices):
                if device.name and device.name == device_name[d]:
                    devices.append(device)
        else:
            devices = self.devices

        for device in devices:
            if get_frame:
                if device.device_type == DeviceType.Emg:
                    count = 0
                    for device_tmp in devices:
                        if device_tmp == device:
                            break
                        if device.device_type == DeviceType.Emg:
                            count += 1
                    device.new_data = self.emg_client[count].read()
                else:
                    count = 0
                    for device_tmp in devices:
                        if device_tmp == device:
                            break
                        if device.device_type == DeviceType.Imu:
                            count += 1
                    device.new_data = self.imu_client[count].read()
            if channel_idx:
                device_data = np.ndarray((len(channel_idx), device.new_data.shape[1]))
                for i, idx in enumerate(channel_idx):
                    device_data[i, :] = device.new_data[idx, :]
            device_data = device_data if channel_idx else device.new_data
            if get_frame:
                device.append_data(device.new_data)
            all_device_data.append(device_data)
            if len(all_device_data) == 1:
                all_device_data = all_device_data[0]
        return all_device_data

    def get_frame(self):
        """
        Get a frame from the interface. This function is used to get data from the interface.
        """
        if not self.is_initialized:
            raise RuntimeError("Client is not initialized. Please call init_client() first.")
        self.get_device_data(get_frame=True)
        return True

    def init_client(self):
        """
        Initialize the client if it's not already done. This function has to be called before getting a frame.
        """
        self.is_initialized = True
        for d, device in enumerate(self.devices):
            if device.device_type == DeviceType.Emg:
                self.emg_client.append(
                    pytrigno.TrignoEMG(
                        channel_range=device.device_range, samples_per_read=device.sample, host=self.address
                    )
                )
                self.emg_client[-1].start()
            elif device.device_type == DeviceType.Imu:
                imu_range = (device.device_range[0] * 9, device.device_range[1] * 9)
                self.imu_client.append(
                    pytrigno.TrignoIM(channel_range=imu_range, samples_per_read=device.sample, host=self.address)
                )
                self.imu_client[-1].start()
            else:
                raise RuntimeError("Device type must be 'emg' or 'imu' with pytrigno.")



"""
This example shows how to disseminate EMG data using biosiglive and a dedicated interface.
First an interface is created, then a device is added to the interface. For EMG data, two interfaces are available:
ViconClient to stream data from a vicon Nexus software via the Vicon Data Stream SDK (https://www.vicon.com/software/datastream-sdk/). The Pytrigno interface for streaming data from the Delsys trigno community SDK (https://delsys.com/sdk/).
If you want to try this example offline, you can use the provided custom interface named 'MyInterface' which you can use as a standard interface.

The EMG device taking as input argument :
- the number of electrodes
- the type of the device (the allowed types are listed in the DeviceType class)
- the name of the device (for the Vicon system, it must be the same name as the device to stream).
- the flow of the device.
- the process method to use. Here, the process_emg method is used.
- any other argument needed for the processing method. (you can see the possible arguments in the documentation of the process_emg method).

If you want to display the data in real time, you can use the biosiglive plot classes, please take a look at the live_plot.py example.
After initializing the interface and the device, the data exchange is done in a loop.
The data must first be received from the source by the get_device_data method.
After that, the data can be used as is or processed using the process() method of the Device class.
In this function you can pass a method if you want to use a different method than the default one and every needed argument for that function as well.
"""

from custom_interface import MyInterface
from biosiglive.gui.plot import LivePlot
from biosiglive import (
    ViconClient,
    RealTimeProcessingMethod,
    PlotType,
)
from time import sleep, time


if __name__ == "__main__":
    try_offline = True

    output_file_path = "trial_x.bio"
    if try_offline:
        interface = MyInterface(system_rate=100, data_path="abd.bio")
    else:
        interface = ViconClient(ip="localhost", system_rate=100)

    n_electrodes = 4
    muscle_names = [
        "Pectoralis major",
        "Deltoid anterior",
        "Deltoid medial",
        "Deltoid posterior",
    ]
    interface.add_device(
        nb_channels=n_electrodes,
        device_type="emg",
        name="emg",
        rate=2000,
        device_data_file_key="emg",
        processing_method=RealTimeProcessingMethod.ProcessEmg,
        moving_average=True,
    )

    emg_plot = LivePlot(
        name="emg", rate=100, plot_type=PlotType.Curve, nb_subplots=n_electrodes, channel_names=muscle_names
    )
    emg_plot.init(plot_windows=500, y_labels="Processed EMG (mV)")

    emg_raw_plot = LivePlot(
        name="emg_raw", rate=100, plot_type=PlotType.Curve, nb_subplots=n_electrodes, channel_names=muscle_names
    )
    emg_raw_plot.init(plot_windows=10000, colors=(255, 0, 0), y_labels="EMG (mV)")

    time_to_sleep = 1 / 100
    count = 0
    while True:
        tic = time()
        raw_emg = interface.get_device_data(device_name="emg")
        emg_proc = interface.devices[0].process()
        emg_plot.update(emg_proc[:, -1:])
        emg_raw_plot.update(raw_emg)
        count += 1
        loop_time = time() - tic
        real_time_to_sleep = time_to_sleep - loop_time
        if real_time_to_sleep > 0:
            sleep(real_time_to_sleep)



"""
This file contains a generic interface class to use for any new implemented class.
"""
from .param import *
from typing import Union
from ..enums import DeviceType, InverseKinematicsMethods, InterfaceType


class GenericInterface:
    """
    Class for generic interfacing.
    """

    def __init__(
        self, ip: str = "127.0.0.1", system_rate: float = 100, interface_type: Union[InterfaceType, str] = None
    ):
        """
        Initialize the generic class.

        Parameters
        ----------
        ip: str
            IP address of the interface.
        system_rate: float
            Rate of the system which record the data.
        interface_type: Union[InterfaceType, str]
            Type of the interface.
        """
        self.ip = ip
        self.system_rate = system_rate
        self.acquisition_rate = None
        self.devices = []
        self.marker_sets = []
        self.is_frame = False
        self.kalman = None
        if isinstance(interface_type, str):
            if interface_type not in [t.value for t in InterfaceType]:
                raise ValueError("The type of interface is not valid.")
            self.interface_type = InterfaceType(interface_type)
        else:
            self.interface_type = interface_type

    def _add_device(
        self,
        nb_channels: int,
        device_type: Union[DeviceType, str] = DeviceType.Emg,
        name: str = None,
        rate: float = 2000,
        device_range: tuple = None,
        processing_method: Union[RealTimeProcessingMethod, OfflineProcessingMethod] = None,
        **kwargs,
    ):
        """
        Add a device to the interface.

        Parameters
        ----------
        name: str
            Name of the device.
        device_type: Union[DeviceType, str]
            Type of the device.
        rate: float
            Rate of the device.
        device_range: tuple
            Range of the device.
        processing_method: Union[RealTimeProcessingMethod, OfflineProcessingMethod]
            Processing method to use to process the device.
        kwargs:
            Keyword arguments for the processing method.
        """
        if name in [device.name for device in self.devices]:
            raise RuntimeError(f"The device '{name}' already exists.")
        if isinstance(device_type, str):
            if device_type not in [t.value for t in DeviceType]:
                raise ValueError("The type of the device is not valid.")
            device_type = DeviceType(device_type)
        device_tmp = Device(device_type, nb_channels, name, rate, self.system_rate)
        device_tmp.device_range = device_range if device_range else (0, nb_channels)
        device_tmp.interface = self.interface_type
        device_tmp.processing_method = processing_method
        device_tmp.processing_method_kwargs = kwargs
        return device_tmp

    def _add_marker_set(
        self,
        nb_markers: int,
        name: str = None,
        marker_names: Union[str, list] = None,
        rate: float = 100,
        unlabeled: bool = False,
        unit: str = "m",
        kinematics_method: Union[InverseKinematicsMethods, str] = None,
        **kwargs,
    ):
        """
        Add a marker set to stream from the interface.

        Parameters
        ----------
        nb_markers: int
            Number of markers in the marker set.
        name: str
            Name of the markers set.
        marker_names: Union[list, str]
            List of markers names.
        rate: int
            Rate of the markers set.
        unlabeled: bool
            Whether the markers set is unlabeled.
        unit: str
            Unit of the markers set.
        kinematics_method: Union[InverseKinematicsMethods, str]
            Inverse kinematics method to use to process the markers set.
        kwargs:
            Keyword arguments for the inverse kinematics method.
        """
        if name in [marker.name for marker in self.marker_sets]:
            raise RuntimeError(f"The marker set '{name}' already exists.")
        markers_tmp = MarkerSet(nb_markers, name, marker_names, rate, unlabeled, self.system_rate)
        markers_tmp.kin_method = kinematics_method
        markers_tmp.kin_method_kwargs = kwargs
        markers_tmp.interface = self.interface_type
        markers_tmp.unit = unit
        return markers_tmp

    def add_marker_set(self, **kwargs):
        """
        Add a marker set to stream from the interface system.
        """
        raise RuntimeError(f"Markers are not implemented with interface '{self.interface_type}'.")

    def get_device(self, name: str = None, idx: int = None):
        """
        Get a device from the interface.

        Parameters
        ----------
        idx: int
            Index of the device.
        name: str
            Name of the device.

        Returns
        -------
        The device object.
        """
        if idx is not None:
            if name:
                raise RuntimeError("You cannot provide an index and a name for the device.")
            return self.devices[idx]
        elif name is not None:
            for device in self.devices:
                if device.name == name:
                    return device
        else:
            raise RuntimeError("You must provide an index or a name for the device.")

    def get_marker_set(self, name: str = None, idx: int = None):
        """
        Get a device from the interface.

        Parameters
        ----------
        name: str
            Name of the device.
        idx: int
            Index of the device.

        Returns
        -------
        The device.
        """
        if idx is not None:
            if name:
                raise RuntimeError("You cannot provide an index and a name for the device.")
            return self.marker_sets[idx]
        elif name is not None:
            for marker_set in self.marker_sets:
                if marker_set.name == name:
                    return marker_set
        else:
            raise RuntimeError("You must provide an index or a name for the marker set.")

    def add_device(self, **kwargs):
        """
        Add a device to the interface.
        """
        raise RuntimeError(f"Devices are not implemented with interface '{self.interface_type}.")

    @staticmethod
    def get_force_plate_data():
        """
        Get the force plate data.
        """
        raise NotImplementedError("Force plate streaming is not implemented yet.")

    def get_device_data(self, **kwargs):
        """
        Get the device data.
        """
        raise RuntimeError(f"You can not get device data from '{self.interface_type}'.")

    def get_marker_set_data(self, **kwargs):
        """
        Get the marker set data.
        """
        raise RuntimeError(f"You can not get markers data from '{self.interface_type}'.")

    def get_latency(self):
        """
        Get the latency of the interface.
        """
        return -1

    def get_frame(self):
        """
        Get the frame of the interface. That need to be call to retrieve all the data once at the same time.
        """
        return True

    def get_frame_number(self):
        """
        Get the frame number of the interface.
        """
        raise RuntimeError(f"You can not get frame number from '{self.interface_type}'.")

    def get_kinematics_from_marker_set(self, **kwargs):
        """
        Get the kinematics from a marker set.
        """
        raise RuntimeError(f"You can not get kinematics from '{self.interface_type}'.")

    def init_client(self):
        """
        Initialize the client.
        """
        raise RuntimeError(f"You can not init client from '{self.interface_type}'.")
